##  Functions for manipulating PHREEQC databases

### Marco De Lucia, delucia@gfz-potsdam.de, 2009-2018
### Time-stamp: "Last modified 2018-05-03 17:23:09 delucia"

##' @title Finds the expression of a logK of a given compound/mineral
##'     in a PHREEQC database
##' @param species name of the species/mineral to search for
##' @param db the buffer containing the database
##' @param type one of c("analytical","logk","equation","lhs","rhs")
##' @return the requested logK
##' @author MDL
##' @export
FindLogK <- function(species, db, type=c("analytical"))
{
    ##  reg <- gsub("\\(","\\\\(",species) ## obsoleted through the fixed=TRUE option
    ##  reg <- gsub("\\)","\\\\)",reg)

    args <- c("analytical","logk","equation","lhs","rhs")
    type <- match.arg(type, args)
    
    pos  <- grep(species,db,fixed=TRUE)
    if (length(pos) > 1) pos <- pos[1] ## just the 1st occurrence
    if (length(pos) == 0) {
        msg("Species \"", species,"\" not found in the DataBase")
        return(NA)
    }
    ## check if analytical expression is present!
    if (type=="analytical") {
        linepos <- grep("anal",db[c(pos:(pos+4))])
        if (length(linepos)==1) {
            
            linepos <- pos + linepos - 1
            line <- TrailSpaces(db[linepos])
            
            parread <- unlist(strsplit(line," +"))[-1]
            len <- length(parread)
            if (len != 5) {
                parread[seq(len+1,5)] <- 0
                msg("Analytical expression for reaction",species,"has only", len, "terms")
            }
            par.anal <- as.numeric(parread)
            attr(par.anal,"line") <- linepos
            attr(par.anal,"type") <- "analytical"
            return(par.anal)
        } 
    }
    if (type=="logk") {
        
        ## if we are here, either there is no analytical, or we want to calculate also the new logK
        linepos <- grep("log",db[c(pos:(pos+3))])
        if (length(linepos)==1) {
            linepos <- pos + linepos - 1
            line <- TrailSpaces(db[linepos])
            
            parread <- unlist(strsplit(line," +"))[2]
            
            par.logk <- as.numeric(parread)
            attr(par.logk,"line") <- linepos
            attr(par.logk,"type") <- "logk"
            
            return(par.logk)
        }
        
    }
    if (type=="equation") {
        return(TrailSpaces(db[pos+1]))
    }
    if (type=="lhs") {
        return(ParseFormula(TrailSpaces(db[pos+1]), type="lhs"))
    }
    if (type=="rhs") {
        return(ParseFormula(TrailSpaces(db[pos+1]), type="rhs"))
    }
    ## if we are here, no matching expression found, returning NA
    return(NA)

}

##' @title Calculates the logK of reaction for a given temperature
##' @param param the vector containing the coefficients of analytical
##'     logK
##' @param Tk temperature in Kelvin
##' @return The logK at wanted temperature
##' @author MDL
##' @export
AnalyticalLogK <- function(param,Tk)
{
  return(param[1]+param[2]*Tk+param[3]/Tk+param[4]*log10(Tk)+param[5]/Tk^2)
}


##' @title Parses a formula as it is written in a PHREEQC database
##' @param line the string containing the reaction
##' @param type one of c("all","lhs","rhs")
##' @return a list containing components and coefficients.
##'     Coefficients of left hand side are negated.
##' @author MDL
##' @export
ParseFormula <- function(line, type="all")
{
    args <- c("all","lhs","rhs")
    type <- match.arg(type, args)

    tsline <- TrailSpaces(line)
    formula <- unlist(strsplit(tsline,"="))
    left <- formula[1]
    right <- formula[2]
  
    lhs <- unlist(strsplit(left," + ", fixed=TRUE))
    rhs <- unlist(strsplit(right," + ", fixed=TRUE))


    if (type=="all" || type=="lhs") {
        ## left hand side
        coeffsl <- rep(1,length(lhs))
        compl <- rep("",length(lhs))
        for (i in seq_along(lhs)) {
            lhs[i] <- sub(' +$', '', lhs[i]) ## remove spaces at the end of string
            spaces <- grep(" ",lhs[i],fixed=TRUE)
            if (length(spaces)==1) {
                tmp <- unlist(strsplit(lhs[i]," ", fixed=TRUE))
                coeffsl[i] <- as.numeric(tmp[1])
                compl[i] <- tmp[2]
            } else {
                compl[i] <- lhs[i]
            }
        }
    }
    
    if (type=="all" || type=="rhs") {
        ## right hand side
        coeffsr <- rep(1,length(rhs))
        compr <- rep("",length(rhs))
        for (i in seq_along(rhs)) {
            rhs[i] <- sub('^ +', '', rhs[i]) ## remove spaces at the beginning
            spaces <- grep(" ",rhs[i],fixed=TRUE)
            if (length(spaces)==1) {
                tmp <- unlist(strsplit(rhs[i]," ", fixed=TRUE))
                coeffsr[i] <- as.numeric(tmp[1])
                compr[i] <- tmp[2]
            } else {
                compr[i] <- rhs[i]
            }
        }
    }

    if (type=="all")
        out <- list(comp=c(compl,compr), coeff=c(-coeffsl,coeffsr))
    if (type=="lhs")
        out <- list(comp=compl, coeff=-coeffsl)
    if (type=="rhs")
        out <- list(comp=compr, coeff=coeffsr)
    
    return(out)
}

##' @title Find a PURE_PHASE in a database and parses its formula
##' @param species name of phase to search for
##' @param db a PHREEQC database (buffer)
##' @return a list with the components of the formula and the
##'     stoichiometric coefficients, plus the corresponding line in
##'     the db as attr(,"line")
##' @author MDL
##' @export
FindPhase <- function(species,db)
{
  phases <- grep("^PHASES",db)
  pos  <- grep(species, db, fixed=TRUE)
  
  if (length(pos) > 1) pos <- pos[1] ## just the 1st occurrence

  line <- db[pos+1]
  out <- ParseFormula(TrailSpaces(line))

  ## substitute the composition with the mineral name, lower case
  out$comp[1] <- tolower(species)
  attr(out,"line") <- pos+1
  return(out)
}

##' This utility function is not meant to be used directly by the user
##' @title Remove spaces at the end and beginning of a line
##' @param line the string to remove trailing spaces from
##' @return reformatted string without blanks at beginning and end
##' @author MDL
##' @export
TrailSpaces <- function(line)
{
  tmp <- sub(' +$', '', line)  ## remove spaces at the end of the lines
  tmp <- sub('^ +', '', tmp)  ## remove spaces at the beginning
  return(tmp)
}

