##  Functions for dealing with simulations with kinetics

### Marco De Lucia, delucia@gfz-potsdam.de, 2009-2018
### Time-stamp: "Last modified 2018-12-12 14:25:38 delucia"

##' This function just runs the generated input buffer - or a list
##' thereof - through \code{phreeqc}. Obviously it requires the
##' \code{phreeqc} package. Also the database must be previously and
##' explicitely loaded by the user using
##' \code{phreeqc::phrLoadDatabaseString(db <-
##' RPhreeFile("phreeqc.dat", is.db=TRUE ))}
##'
##' Currently parallelization is achieved using
##' \code{parallel::parLapply}, this can change in the future. At the
##' moment the function doesn't check that the SELCTED_OUTPUT block is
##' well-formed, so this is responsibility of the user.
##' @title Run the generated input buffer using phreeqc
##' @param input The input buffer, or list of inputs
##' @param procs the number of processes to employ
##' @param second logical, if TRUE skips each odd row
##' @return a data.frame containing the SELECTED_OUTPUT, ordered if
##'     multiple parallel simulations were called. Some manipulations
##'     of the colnames are also made.
##' @author MDL
##' @export
RunPQC <- function(input, procs=1, second=TRUE) {

    .runPQC <- function(input, onlysecond) {
        require(phreeqc)
        phreeqc::phrRunString(input)
        out <- phreeqc::phrGetSelectedOutput()[[1]]
        nso <- colnames(out)
        nso <- sub('..mol.kgw.','',nso, fixed=TRUE)
        nso <- sub('.mol.kgw.','',nso, fixed=TRUE)
        nso <- sub('^k_','',nso)
        nso <- sub('.g.','(g)',nso, fixed=TRUE)
        if (second)
            out <- out[seq(2, nrow(out), by=2),]
        colnames(out) <- nso
        return(data.matrix(out))
    }

    ## is lin a list of inputs?
    if (!is.list(input)) {
        if(is.character(input)) { ## normal sequential run
            res <- .runPQC(input, onlysecond=second)
        } else {
            stopmsg("something wrong with the input, dying!")
        }
    } else {
        ## go parallel!

        ## old one
        ## res <- parallel::mclapply(input, .runPQC, mc.silent=TRUE, mc.cores=procs)
        ## res <- parallel::parLapply(ThisRunCluster, input, .runPQC)
        res <- foreach(i=seq_along(input), .combine=rbind) %dopar% .runPQC(input[[i]], onlysecond=second)

        ## ## a is the string containing the rbind of each element of the list
        ## a <- paste("rbind(",paste("res[[",1:procs,"]]",sep="", collapse = ", "),")")
        ## ## let evaluate this expression
        ## res <- eval(parse(text=a))
    }
    return(res)
}

##' Function to distribute different values of one property
##' (concentration, temperature, mineral phase,\dots) properties in an
##' input buffer, special kinetics case.
##'
##' The number of SOLUTIONs in the input buffer must be 1 or equal to
##' length(values). If the input buffer contains only one SOLUTION,
##' and length(values)>1, then the new input gets automatically
##' replicated length(values) times by RepSol.
##' @title Distribute properties in an input buffer, KINETICS case
##' @param input The initial input buffer.
##' @param prop The property whose values need to be distributed.
##' @param values The numerical value(s) to distribute across the
##'     SOLUTIONS/KINETICS blocks.
##' @param ident The identifier for the line where the actual values will be pasted, defaults to "-m"
##' @return A new input buffer
##' @author MDL
##' @export
DistributeKin <- function(input, prop=NULL, values, ident="-m0")
{
    if(is.null(prop))
        return(Distribute(input, prop=ident, values=values))
    
    ## correct quoting of parenthesis
    reg <- gsub("\\(","\\\\(",prop)
    reg <- gsub("\\)","\\\\)",reg)

    ## coerce values to character
    n <- length(val <- as.character(values))
    newinp <- input
    ## find the offset between "kinetics" and "end" statements
    initline <- grep("^KINETIC", newinp)
    endline <- grep("^END", newinp)

    ## this is the first kinetic block
    kinblock <- newinp[seq(initline[1],endline[1])]

    ## find which kinetic mineral inside the KINETICS block
    kinblock_str <- kinblock[-grep("^-",kinblock)]  # strip all lines starting with "-"
    kinblock_str <- kinblock_str[-c(1, length(kinblock_str))] # remove "KINETICS" and "END"
    offset_kin <- grep(reg, kinblock_str, fixed=TRUE) ## this is the position of the mineral inside a block

    ## position of the identifier inside kinetic block
    idline <- grep(paste0(ident," "), kinblock, fixed=TRUE)[offset_kin] 

    ## assuming all kinetics blocks are the same, add this offset to the vector of "KINETICS"
    ids <- idline + initline -1

    if (length(ids)!=n)
        stopmsg("Length must match the number of values")
    
    newinp[ids] <- paste(ident,values,sep="  ")
    return(newinp)  
}

##' TODO
##'
##' TODO
##' @title Splits a named matrix of data (a column is a variable) in n simulations to run in parallel
##' @param data The matrix or data.frame containing the values to be splitted
##' @param procs How many processors?
##' @param base The base script SOLUTION block
##' @param first The first block to be prepended to each generated input
##' @param prop The names of the properties
##' @param minerals vector containing the indeces of the EQUILIBRIUM minerals
##' @param kin vector containing indeces of KINETICS minerals
##' @param dt time step
##' @param ann optional annotation to be postponed to the values
##' @param nmax minimum number of simulations in one "split"
##' @param verbose logical, defaulting to FALSE. If TRUE, some message are displayed
##' @return a list of length "procs" with the inputs, ready to be run in parallel
##' @author MDL
##' @export
splitMultiKin <- function(data, procs, base, first, prop, minerals, kin, dt, ann, nmax=200, verbose=FALSE)
{
    if (is.matrix(data))
        ntot <- dim(data)[1] ## the total number of pqc simulations
    else {
        datamat <- matrix(data,ncol=length(prop),byrow=TRUE)
        colnames(datamat) <- prop
        data <- datamat
        ntot <- nrow(datamat)
    }

    if (!missing(ann)) {
        annlen <- length(ann)
        annprop <- names(ann)
    }

    if ( ntot==1) {
        ## just 1 simulation!)
        tmpdata <- as.numeric(data)
        indAq <- seq(1:length(prop))[-minerals]
        newinp <- base
        
        for (i in indAq) {
            newinp <- Distribute(newinp, prop[i], tmpdata[i])
        }

        for (i in setdiff(minerals, kin)) {

            ## if we give a further annotation as SI
            if (prop[i] %in% annprop) {
                indann <- which(annprop==prop[i])
                newinp <- Distribute(newinp, prop[i], paste(ann[[indann]], tmpdata[i]))
            } else {
                newinp <- Distribute(newinp, prop[i], paste("0.0 ",tmpdata[i]))
            }
        }

        for (i in kin) {
            newinp <- DistributeKin(newinp, prop[i], tmpdata[i], ident="-m")
        }

        ## And finally the time
        newinp <- Distribute(newinp, "-steps", dt)
        return(c(first,newinp))

    }
 
    if (ntot > max(nmax,procs*2) && procs > 1) {
        if (verbose)
            msg(" Splitting up ", ntot, "simulations to", procs, "processes...")
        ## find best division in n
        nsim <- ntot%/%procs
        nlast <- nsim + ntot%%procs
        totl <- c(rep(nsim,procs-1), nlast)
        breaks <- c(0,cumsum(totl))
        
        BigInpLast <- RepSol(base, nlast) 
        BigInp     <- RepSol(base, nsim)
        
        datalist  <- vector(mode="list", length=procs)
        inputlist <- vector(mode="list", length=procs)
        for (i in seq_along(datalist))  {
            datalist[[i]] <- data[c( (breaks[i]+1):breaks[i+1] ) , ]
            inputlist[[i]] <- BigInp
        }
        inputlist[[length(inputlist)]] <- BigInpLast
        ## attr(input,"sims") <- breaks
        
        for (i in seq_along(datalist)) {
            inputlist[[i]] <- DistributeKinMatrix(inputlist[[i]], prop, datalist[[i]], minerals=minerals,
                                                  kin=kin, dt=dt, ann=ann)
            inputlist[[i]] <- c(first,inputlist[[i]])
        }
        return(inputlist)

    } else  {
        if(verbose)
            msg("won't parallelize... procs=", procs," ntot=", ntot)
        BigInp <- RepSol(base, ntot)
        tmp <- DistributeKinMatrix(input=BigInp, prop=prop, values=data, minerals=minerals, kin=kin, dt=dt, ann=ann)
        input <- c(first, tmp)
        return(input)
    }
    ## if we arrive here must be gone something wrong
    stopmsg("BUMMER!!! Something wrong in here, shutting down")
}


##' @title Distribute all values from a matrix into a buffer, EQUILIBRIUM case
##' @param input The input buffer to manipulate
##' @param prop  The names of properties to distribute
##' @param values matrix or data.frame containing the values to distribute
##' @param minerals indeces of the EQUILIBRIUM minerals
##' @param SI logical, defaults to FALSE. Should we paste SI before value of a mineral?
##' @return input buffer with the values in correct places
##' @author MDL
##' @export
DistributeMatrix <- function(input, prop, values, minerals, SI=FALSE)
### Distribute properties in an input buffer
### prop is a matrix with colnames
{
    if (!is.matrix(values)&&!is.data.frame(values))
        stopmsg("I need a matrix to distribute")
    ## correct quoting
    qprop <- gsub("\\(","\\\\(",prop)
    qprop <- gsub("\\)","\\\\)",qprop)
    n <- length(qprop)
    if (n!=ncol(values)) {
        stopmsg("prop names and values matrix of different dimension.")
    }
    if (is.character(minerals))
        {
            qmin <- gsub("\\(","\\\\(",minerals)
            qmin <- gsub("\\)","\\\\)",qmin)
            mins <- match(qmin, qprop)
        }
    else
        mins <- as.numeric(minerals)
    
    indAq <- seq(1:n)[-mins]
    newinp <- input

    for (i in indAq) {
            newinp <- Distribute(newinp, prop[i], values[,i])
    }

    for (i in mins) {
        if (SI)
            newinp <- Distribute(newinp, prop[i], values[,i])
        else
            newinp <- Distribute(newinp, prop[i], paste("0.0 ",values[,i]))
    }

    return(newinp)
}

##' This function \link{DistributeKin} a matrix of properties values into a buffer
##'
##' TODO
##' @title Distribute all values from a matrix into a buffer, KINETICS case
##' @param input The input buffer to manipulate
##' @param prop The names of properties to distribute
##' @param values matrix or data.frame containing the values to distribute
##' @param minerals indeces of the EQUILIBRIUM minerals
##' @param kin indeces of KINETICS minerals
##' @param dt Time step
##' @param ann optional annotations to be postponed to the values
##' @return input buffer with the values in correct places
##' @author MDL
##' @export
DistributeKinMatrix <- function(input, prop, values, minerals, kin, dt, ann)
{
    if (!is.matrix(values)&&!is.data.frame(values))
        stop("I need a matrix to distribute")
    ## correct quoting
    qprop <- gsub("\\(","\\\\(",prop)
    qprop <- gsub("\\)","\\\\)",qprop)
    n <- length(qprop)
    if (n!=ncol(values)) {
        msg(" BUMMER!")
        stopmsg("prop names and values matrix of different dimensions?")
    }

    if (is.character(minerals)) {
        qmin <- gsub("\\(","\\\\(",minerals)
        qmin <- gsub("\\)","\\\\)",qmin)
        mins <- match(qmin, qprop)
    } else {
        mins <- as.integer(minerals)
    }
    if (is.character(kin)) {
        qkin <- gsub("\\(","\\\\(",kin)
        qkin <- gsub("\\)","\\\\)",qkin)
        kins <- match(qkin, qprop)
    } else {
        kins <- as.integer(kin)
    }
    
    indAq <- seq(1:n)[-mins]
    newinp <- input


    do_ann <- FALSE
    if (is.list(ann)) {
        annlen <- length(ann)
        annprop <- names(ann)
        do_ann <- TRUE
    }
    
    for (i in indAq) {
        newinp <- Distribute(newinp, prop[i], values[,i])
    }
    for (i in setdiff(mins, kin)) {
        if (do_ann) {
            if (prop[i] %in% annprop) {
                indann <- which(annprop==prop[i])
                newinp <- Distribute(newinp, prop[i], paste(ann[[indann]], values[,i]))
            }
        } else {
            newinp <- Distribute(newinp, prop[i], paste("0.0 ", values[,i]))
        }
    }
    for (i in kin) {
        newinp <- DistributeKin(newinp, prop[i], values[,i], ident="-m")
    }

    ## And finally the time
    newinp <- Distribute(newinp, "-steps", rep(dt, nrow(values)))
    
    return(newinp)
}

##' @title Splits a data (matrix) in n simulations to run in parallel
##' @param data the data matrix
##' @param procs number of processors to split to
##' @param base base input script
##' @param first block to be prepended to each manipulated script
##' @param prop vector containing the names of the properties
##' @param minerals indeces of EQUILIBRIUM minerals
##' @param nmax max number of simulations in each input
##' @param verbose logical, defaulting to FALSE. If TRUE, some message are displayed
##' @return a list with n=procs inputs
##' @author MDL
##' @export
splitMultiFix <- function(data, procs, base, first, prop, minerals, nmax=200, verbose=FALSE)
{
    if (is.matrix(data))
        ntot <- dim(data)[1] ## the total number of pqc simulations
    else {
        datamat <- matrix(data,ncol=length(prop),byrow=TRUE)
        colnames(datamat) <- prop
        data <- datamat
        ntot <- nrow(datamat)
    }
    
    if ( ntot  ==1) ## try and call normal Distribute (i.e., ##!is.matrix(data) ||
        ## just 1 simulation!)
    {
        if (verbose)
            msg("just 1 simulation, reverting back to normal Distribute")
        BigInp <- RepSol(base, 1, first=first)
        namdis <- colnames(data)
        indminer <- minerals
        nammin <- namdis[indminer]
        namtot <- namdis[-indminer]
        for (dis in seq_along(nammin)) {
            BigInp <- Distribute(BigInp, nammin[dis], paste("0.0 ", data[1,nammin[dis] ]))
        }
        for (dis in seq_along(namtot)) {
            BigInp <- Distribute(BigInp, namtot[dis], data[1,namtot[dis] ])
        }
        return(BigInp)
    }
 
    if (ntot > nmax && procs > 1)   {
        if (verbose)
            msg("Splitting up ", ntot, "simulations to", procs, "processes...")
        ## find best division in n
        nsim <- ntot%/%procs
        nlong <- nsim + ntot%%procs
        totl <- c(nlong,rep(nsim,procs-1))
        breaks <- c(0,cumsum(totl))
        
        BigInplong <- RepSol(base, nlong, first=first)
        BigInp <- RepSol(base, nsim, first=first)
        
        newlist <- vector(mode="list", length=procs)
        for (i in seq_along(newlist)) {
            newlist[[i]] <- data[c( (breaks[i]+1):breaks[i+1] ) , ] 
        }
        
        input <- vector(mode="list", length=procs)
        attr(input,"sims") <- breaks
        
        input[[1]] <- DistributeMatrix(BigInplong, prop, newlist[[1]], minerals=minerals)
        input[c(2:procs)] <- lapply(newlist[c(2:procs)],
                                    function(x)
                                        return(DistributeMatrix(BigInp, prop, x, minerals=minerals))
                                    )
    }  else {
        BigInp <- RepSol(base, ntot, first=first)
        input <- DistributeMatrix(input=BigInp, prop=prop, values=data, minerals=minerals)
    }
    return(input)
}

