##  Functions for dealing with surrogate simulations

### Marco De Lucia, delucia@gfz-potsdam.de, 2009-2018
### Time-stamp: "Last modified 2019-05-08 17:19:57 delucia"


##' Todo
##'
##' @title Simple explicit 1D forward-Euler advection, multispecies
##' @param conc matrix containing the chemical state
##' @param inflow vector of concentrations entering the inlet
##' @param dx the grid spacing
##' @param dt the required time step
##' @return a matrix containing the state after the advection
##' @author MDL
##' @export
AdvectionPQC <- function(conc, inflow=rep(0,ncol(conc)), dx, dt)
{
    if (is.matrix(conc)) {
        ##        msg("conc is matrix")
        ## CurrConc <<- conc
        ## actually the first iteration does not have the attributes attached
        if ( "immobile" %in% names(attributes(conc)) ) { 
            immobile <- attr(conc,"immobile")
            val <- conc[,-immobile]
        } else {
            msg("input does not have immobile columns!")
            val <- conc
        }
        ## sort the columns of the matrix matching the names of boundary 
        sptr <- colnames(val)
        spin <- names(inflow) 
        ind <- match(spin,sptr)
        if (TRUE %in% is.na(ind)) {
            val <- cbind(val,matrix(0,ncol=sum(is.na(ind)),nrow=nrow(val) ))
            colnames(val) <- c(sptr,spin[is.na(ind)])
        }

        sptr <- colnames(val)
        ind <- match(sptr,spin)
        bound <- numeric(ncol(val))
        bound[!is.na(ind)] <- inflow[na.omit(ind)]
        cj <- rbind(bound,val)
        remnames <- colnames(cj)
        if (typeof(cj)!="double") {
            msg("cj is not 'double'")
            ## mode(cj) <- "double"
        }
        cnew <- -U*dt/dx*diff(cj) + val 

        ## if the original input had immobile attribute and columns
        ## then reattach the columns and the attribute
        if ( "immobile" %in% names(attributes(conc)) ) { 
            res <- cbind(cnew, conc[,immobile])
            
            attr(res,"immobile") <- immobile 
        } else
            res <- cnew

        colnames(res) <- colnames(conc)

        if (typeof(res)!="double") {
            ##mode(res) <- "double"
            msg("res is not 'double'")
        }

        return(res)
    } else {
        cj <- c(inflow,conc)
        cnew <- -U*dt/dx*diff(cj)+conc
        if (typeof(cnew)!="double") {
            msg("cnew is not double")
            ## mode(cnew) <- "double"
        }

        return(cnew)
    }
}

##' Makes sure that we get activities instead of pH/pe
##'
##' The manipulation of the state should be in place
##' @title Ensure that pH/pe are transformed into activities
##' @param state a chemical state (coerced to data.matrix)
##' @return the state where, if necessary, pH is transformed into activity(H+)
##' @author MDL
##' @export
pH2Act <- function(state){
    rem_attr <- attributes(state)
    
    stateNames <- colnames(state)
    if (mode(state)!="double" & length(dim(state)) ==2) {
        mat <- apply(state, 2, as.numeric)
    } else {
        mat <- data.matrix(state)
    }

    if ( "pH" %in% stateNames) {
        mat[,"pH"] <- 10^(-mat[,"pH"])
        stateNames[stateNames=="pH"] <- "H"
   }
    if ( "pe" %in% stateNames) { 
        mat[,"pe"] <- 10^(-mat[,"pe"])
        stateNames[stateNames=="pe"] <- "e"
    }

    if ("immobile" %in% names(rem_attr)) attr(mat,"immobile") <- rem_attr$immobile
    if ("index" %in% names(rem_attr)) attr(mat,"index") <- rem_attr$index
    colnames(mat) <- stateNames
    return(mat)
}

##' Makes sure that we get pH/pe instead of activities
##'
##' The manipulation of the state should be in place
##' @title Ensure that activities are transformed into pH/pe
##' @param state a chemical state (coerced to data.matrix)
##' @return the state where, if necessary, activity(H+) is transformed
##'     into pH
##' @author MDL
##' @export
Act2pH <- function(state){
    rem_attr <- attributes(state)

    stateNames <- colnames(state)

    if (mode(state)!="double" & length(dim(state)) ==2) {
        mat <- apply(state, 2, as.numeric)
    } else {
        mat <- data.matrix(state)
    }
    if ( "H" %in% stateNames) {
        mat[,"H"] <- -log10(mat[,"H"])
        stateNames[stateNames=="H"] <- "pH"
    }
    if ( "e" %in% stateNames) { 
        mat[,"e"] <- -log10(mat[,"e"])
        stateNames[stateNames=="e"] <- "pe"
    }

    if ("immobile" %in% names(rem_attr)) attr(mat,"immobile") <- rem_attr$immobile
    if ("index" %in% names(rem_attr)) attr(mat,"index") <- rem_attr$index
    colnames(mat) <- stateNames
    return(mat)
}


##' TODO
##'
##' @title Recompose a chemical state after reduction
##' @param state the state to be recomposed
##' @param reduced the initial reduced state containing the "index" attribute
##' @return a state with the original (full-grid) number of elements
##' @author MDL
##' @export
RecomposeState <- function(state, reduced)
{
    if ("index" %in% names(attributes(reduced))) {
        inds <- attr(reduced,"index")
        ## msg("print state:")
        ## print(state)
        ## msg("print inds:")
        ## print(inds)
    } else stopmsg("no reduction indeces!")
    res <- state[inds,]
    attr(res,"immobile") <- attr(reduced,"immobile")
    return(res)
}

##' This functions requires the package \code{mgcv}
##'
##' TODO
##' @title Finds unique problems in a geochemical state
##' @param data the matrix
##' @return a matrix plus an attribute containing the indeces of the
##'     compressed matrix
##' @author MDL
##' @export
ReduceState <- function(data)
{
    require(mgcv)
    
    red <- mgcv::uniquecombs(data)
    colnames(red) <- colnames(data)
    attr(red,"immobile") <- attr(data,"immobile") 
    return(red)
}



##' TODO
##' @title 1D Reactive transport simulations using phreeqc or a
##'     surrogate, equilibrium simulations
##' @param setup a list with several input parameters
##' @param init optional, initial state of the simulation
##' @param maxtime depending on \code{step}, the simulation time
##' @param step one of c("time","iter","fix_dt"), defines the time
##'     stepping
##' @param db the chemical database
##' @param procs number of processors to be used in parallel execution
##' @param maxsim number of simulations above which parallelization is
##'     triggered
##' @param reduce logical, defaults to TRUE. Should we find the unique
##'     occurences of a geochemical problem?
##' @param ebreak logical, defaults to FALSE. If TRUE, an early break
##'     is invoked
##' @param writeout logical, defaults to FALSE.
##' @param surrogate logical, defaults to FALSE. Should we use
##'     surrogates instead of phreeqc?
##' @param surrogate.FUN The function which calls the surrogate
##' @param model The list of surrogate models, one element per
##'     variable
##' @param baleq data structure containing the balance equations
##' @param tol absolute tolerance on the mass balance
##' @param call_pqc logical, defaults to TRUE: if TRUE, calls phreeqc
##'     on the surrogate simulations which trespass the tolerance
##' @return a list containing lots of stuff
##' @author MDL
##' @export 
ReactTranspBalanceEq <- function(setup, init, maxtime, step=c("time","iter","fix_dt"),
                                 db="phreeqc.dat", procs=1, maxsim=150,
                                 reduce=TRUE, ebreak=FALSE, writeout=FALSE, surrogate=FALSE,
                                 surrogate.FUN, model, baleq, tol=1E-9, call_pqc=TRUE)
    
{
    require(doParallel)
    if (surrogate) {
        if (missing(surrogate.FUN))
            stopmsg("Need a valid function to apply surrogate instead of chemistry")
        if (is.function(surrogate.FUN))
            Selected.Surrogate <- surrogate.FUN
    }
    
    step <- match.arg(step)
    ## prototype of the PHREEQC simulation
    base     <- setup$base
    ## SELECTED_OUTPUT and USER_PUNCH sections of the PHREEQC input data
    first    <- setup$first
    ## boundary conditions
    bound    <- setup$bound
    ## prop are the names of the properties for PHREEQC
    prop     <- setup$prop
    immobile <- setup$immobile
    initsim  <- setup$initsim

    if (procs > 1) {
        if (Sys.info()[["sysname"]]=="Windows") {
            ThisRunCluster <<- parallel::makePSOCKcluster(procs)
        } else {
            ThisRunCluster <<- parallel::makeForkCluster(procs)
        }

        doParallel::registerDoParallel(ThisRunCluster)
        msg("Registered default doParallel cluster with ", procs, "nodes")
    }

    n  <- setup$n
    L  <- setup$len  # m
    U  <- setup$U    # m3/s
    dx <- L/n        # m
    dt <- dx/U * setup$Cf
    msg("Pure advection 1D simulations")

    ## Find out the number of iterations we're about to calculate
    if (step=="fix_dt"){
        dt <- setup$dt
        maxiter <- floor(maxtime/dt)+1
        msg("Fixed time step of ", dt, ". ", maxiter, " iterations required to reach enditme of", maxtime)
    }
    
    if (step=="time") {
        maxiter <- floor(maxtime/dt)+1
        msg("Courant number=", U*dt/dx, "; dt=",dt)
        msg(maxiter,"iterations required") 
    }
    
    if (step=="iter") {
        maxiter <- maxtime
        msg("Courant number=", U*dt/dx, "; dt=",dt,";")
        msg("Simulation will end at approx ",floor(maxiter*dt),"secs") 
    }
    
    ## MDL: add "complete" list. This will be a list of lists. Each
    ## iteration is a 3-components sublist storing state_T (the input
    ## to chemistry, after transport step), state_C (the input to
    ## transport, after chemistry), and the whole PHREEQC output.
    saved_complete <- vector(mode="list",length=maxiter+1)
    pad <- floor(log10(maxiter+1))+1 ## to properly format the step names
    spr_string <- paste0("%0", pad,"d")
    names(saved_complete) <- paste0("step_",sprintf(spr_string,seq(0,maxiter)))
    
    out_inp <<- vector(mode="list",length=length(maxiter))
    out_res <<- vector(mode="list",length=length(maxiter))
    out_bal <<- vector(mode="list",length=length(maxiter))
    out_of_tol <<- vector(mode="list",length=length(maxiter))

    timing <- matrix(NA,ncol=5,nrow=maxiter)

    msg("Loading phreeqc db... ")
    phreeqc::phrLoadDatabaseString(db)
    msg("database loaded.")

    if (procs > 1) {
        parallel::clusterCall(cl=ThisRunCluster, phreeqc::phrLoadDatabaseString, db)
        msg("Database loaded on each worker.")
        if (surrogate) {
            parallel::clusterExport(cl=ThisRunCluster, varlist=c("model","surrogate.FUN"), envir = environment())
            msg("All workers are setup with the surrogate model.")
        }
    }

    if (missing(init)) {
        msg("missing initial chemical state")
        if (!is.character(initsim))
            runinitsim <- c(first,base)
        else
            runinitsim <- initsim
        
        msg("running initsim PHREEQC script assuming EQUILIBRIUM and correct PUNCH specifications...")
        tmpfirstrun <- RunPQC(runinitsim)
        msg("assuming homogeneous medium")
        msg("Names gathered after first run:", paste(colnames(tmpfirstrun), collapse=", "))
        state_C <- matrix(rep(tmpfirstrun,n), byrow=TRUE, ncol=length(prop))
        colnames(state_C) <- colnames(tmpfirstrun)
    } else {
        msg("given initial state; assuming correct colnames")
        state_C <- init
    }
    
    attr(state_C,"immobile") <- immobile

    ## ssC <<- state_C
    index_saved <- 1
    saved_complete[[index_saved]] <- list(C=Act2pH(state_C))
    index_saved <- index_saved + 1
    
    ## transform the pH/pe back into activities for Advection
    msg("first step, Advection()...")
    state_T <- AdvectionPQC(pH2Act(state_C), inflow=bound, dx=dx, dt=dt)
    msg("Advection done...")
    Tr <- Ts <- 0
    
    ## treat special H+/pH, e-/pe cases
    state_T <- Act2pH(state_T)
    
    if (reduce)
        Tr <- system.time( reduced <- ReduceState(state_T) )[3]
    else
        reduced <- state_T
    
    msg("first step, chemistry ...")
    Ts <- system.time(inplist <- splitMultiFix(data=reduced, procs=procs, base=base, first=first,
                                               prop=prop, minerals=immobile, nmax=maxsim))[3]
    
    Tm <- system.time(res_C <- RunPQC(inplist, procs=procs))[3]
    attr(res_C,"immobile") <- immobile
    
    msg("iteration 0, CPU-time = ",round(Tm,3),"[s]")

    ## recompose after the reduction
    if (reduce)
        Tr <- Tr + system.time(state_C <- RecomposeState(Act2pH(res_C), reduced))[3]
    else {
        state_C <- res_C
        immobile_col_names <- setup$prop[setup$immobile]
        immobile_inds <- match( immobile_col_names, colnames(state_C) )
        attr(state_C,"immobile") <- immobile_inds
    }
    
    saved_complete[[index_saved]] <- list(T=Act2pH(state_T), C=Act2pH(state_C))
    index_saved <- index_saved + 1
    out_inp[[1]] <<- inplist
    out_res[[1]] <<- res_C
    
    if (ebreak) {
        msg("Early break invoked, bye.")
        return(saved_complete[[1]])
    }
    time <- dt
    iter <- 1

    timing[iter,] <- c(dim(reduced)[1],procs,Tm, Tr, Ts)

    msg("going through iterations now!")
    
    on.exit({
        msg("MAIN LOOP interrupted during iteration",iter,"!!")
        msg("returning last calculated chemistry.")
        print(traceback())
        msg(" Bye.")
        attr(state_C,"timing") <- timing
        saved_complete$currentT <- Act2pH(state_T)
        saved_complete$current <- Act2pH(state_C)
        return(saved_complete)
    })
    
    ## start of the iterations
    while (iter<maxiter) {
        ## iterations
        # TODO: It seems to me that iteration numberins is off by one, 
        # so that the first run on Rphree is actually one and the next ones end
        time <- time + dt
        iter <- iter + 1
        #cat("ENTERING ADV","\n")

        ## transform the pH/pe back into activities
        state_T <- AdvectionPQC(pH2Act(state_C), inflow=bound, dx=dx, dt=dt)

        ## treat special H+/pH, e-/pe cases
        state_T <- Act2pH(state_T)

        ## reduction of the problem
        if(reduce)
            Tr <- system.time(reduced <- ReduceState(state_T))[3]
        else
            reduced <- state_T

        ## this is the place where we switch the simulator to the surrogate model
        ## for final test for the geochemical surrogate model!
        
        if (surrogate) { ## use a surrogate model created by the Surrogate playground script
            Ts <- 0.0        ## no split and distribute for this guy

            Tm <- system.time( tmpsur <- Selected.Surrogate(reduced, model=model) )[3]
            bal <- CheckBalance(baleq,reduced,tmpsur)
            out_bal[[iter]] <<- bal
            totbal <- apply(bal, 1, mae)

            nonok <- which(totbal>tol) ### max(tol,tol2))
            if (length(nonok)>0){
                msg("found", length(nonok), "/", nrow(reduced), " offending simulations",
                    ifelse(call_pqc, ", calling PHREEQC"," "))
                if (call_pqc) {
                    reducednonok <- reduced[nonok, ]
                    Ts <- Ts + system.time(inplist <- splitMultiFix(data=reducednonok, procs=procs, base=base, first=first,
                                                                    prop=prop, minerals=immobile, nmax=maxsim))[3]
                    
                    out_inp[[iter]] <- inplist
                    if (length(nonok)==1) {
                        Tm <- Tm + system.time(pqcres_C <- RunPQC(inplist, procs=1))[3]
                    } else {
                        Tm <- Tm + system.time(pqcres_C <- RunPQC(input=inplist, procs=procs))[3]
                    }
                    ##mat <- apply(state, 2, as.numeric)

                    tmpsur[nonok, ] <- pqcres_C
                }
            }
            res_C <- tmpsur
            attr(res_C,"immobile") <- immobile
        } else { ## no surrogate, the original implementation
            ## new input
            Ts <- system.time(inplist <- splitMultiFix(data=reduced, procs=procs, base=base, first=first,
                                                       prop=prop, minerals=immobile, nmax=maxsim))[3]
            out_inp[[iter]] <<- inplist
            Tm <- system.time(res_C <- RunPQC(inplist, procs=procs))[3]
            attr(res_C,"immobile") <- immobile
        }
        
        ## recompose after the reduction
        if (reduce)
            Tr <- Tr + system.time(state_C <- RecomposeState(res_C, reduced))[3]
        else {
            state_C <- res_C
            immobile_col_names <- setup$prop[setup$immobile]
            immobile_inds <- match( immobile_col_names, colnames(state_C) )
            attr(state_C,"immobile") <- immobile_inds
        }
        
        ## ## transform the pH/pe back into activities
        ## state_C <- pH2Act(state_C)

        out_res[[iter]] <<- res_C

        timing[iter,] <- c( dim(reduced)[1], procs, Tm, Tr, Ts)
        msg("done iteration", iter, "/", maxiter," CPU-time ",round(Tm,3),"[s]")

        saved_complete[[index_saved]] <- list(T=Act2pH(state_T), C=Act2pH(state_C))
        index_saved <- index_saved + 1
        
    }
    on.exit()
    
    if (procs > 1) {
        stopCluster(ThisRunCluster)
        msg("Parallelization cluster stopped")
    }


    msg("total time of chemistry: ",round(sum(timing[,3]),3),"seconds")
    msg("total number of simulations: ",sum(timing[,1]))
    msg(" Bye.")
    
    attr(saved_complete,"timing") <- timing
    return(saved_complete)
}


##' Please look at the provided example
##' @title 1D Reactive transport simulations using phreeqc or a
##'     surrogate
##' @param setup a list with several input parameters
##' @param init optional, initial state of the simulation
##' @param maxtime depending on \code{step}, the simulation time
##' @param step one of c("time","iter","fix_dt"), defines the time
##'     stepping
##' @param db the chemical database
##' @param procs number of processors to be used in parallel execution
##' @param maxsim number of simulations above which parallelization is
##'     triggered
##' @param reduce logical, defaults to TRUE. Should we find the unique
##'     occurences of a geochemical problem?
##' @param ebreak logical, defaults to FALSE. If TRUE, an early break
##'     is invoked
##' @param writeout logical, defaults to FALSE.
##' @param surrogate logical, defaults to FALSE. Should we use
##'     surrogates instead of phreeqc?
##' @param surrogate.FUN The function which calls the surrogate
##' @param model The list of surrogate models, one element per
##'     variable
##' @param baleq data structure containing the balance equations
##' @param tol absolute tolerance on the mass balance
##' @param call_pqc logical, defaults to TRUE: if TRUE, calls phreeqc
##'     on the surrogate simulations which trespass the tolerance
##' @return a list containing lots of stuff
##' @author MDL
##' @export 
ReactTranspBalanceKin <- function(setup, init, maxtime, step=c("time","iter","fix_dt"),
                                  db="phreeqc.dat", procs=1, maxsim=150,
                                  reduce=TRUE, ebreak=FALSE, writeout=FALSE, surrogate=FALSE,
                                  surrogate.FUN, model, baleq, tol=1E-9, call_pqc=TRUE)
    
{

    if (surrogate) {
        if (missing(surrogate.FUN))
            stopmsg("Need a valid function to apply surrogate instead of chemistry")
        if (is.function(surrogate.FUN))
            Selected.Surrogate <- surrogate.FUN
    }
    step <- match.arg(step)
    ## prototype of the PHREEQC simulation
    base     <- setup$base
    ## SELECTED_OUTPUT and USER_PUNCH sections of the PHREEQC input data
    first    <- setup$first
    ## boundary conditions
    bound    <- setup$bound
    ## prop are the names of the properties for PHREEQC
    prop     <- setup$prop
    immobile <- setup$immobile
    kin      <- setup$kin
    initsim  <- setup$initsim
    ann      <- setup$ann


    if (procs > 1) {
        if (Sys.info()[["sysname"]]=="Windows") {
            ThisRunCluster <<- parallel::makePSOCKcluster(procs)
        } else {
            ThisRunCluster <<- parallel::makeForkCluster(procs)
        }

        doParallel::registerDoParallel(ThisRunCluster)
        msg("Registered default doParallel cluster with ", procs, "nodes")
    }

    n  <- setup$n
    L  <- setup$len  # m
    U  <- setup$U    # m3/s
    dx <- L/n        # m
    dt <- floor(dx/U * setup$Cf)
    msg("convective simulations")
    
    ## Find out the number of iterations we're about to calculate
    if (step=="fix_dt"){
        dt <- setup$dt
        maxiter <- floor(maxtime/dt)+1
        msg("Fixed time step of ", dt, ". ", maxiter, " iterations required to reach enditme of", maxtime)
    }
        
    if (step=="time") {
        maxiter <- floor(maxtime/dt)+1
        msg("Courant number=", U*dt/dx, "; dt=",dt)
        msg(maxiter,"iterations required") 
    }

    if (step=="iter") {
        maxiter <- maxtime
        msg("Courant number=", U*dt/dx, "; dt=",dt,";")
        msg("Simulation will end at approx ",floor(maxiter*dt),"secs") 
    }

    ## MDL: add "complete" list. This will be a list of lists. Each
    ## iteration is a 3-components sublist storing state_T (the input
    ## to chemistry, after transport step), state_C (the input to
    ## transport, after chemistry), and the whole PHREEQC output.
    saved_complete <- vector(mode="list",length=maxiter+1)
    pad <- floor(log10(maxiter+1))+1 ## to properly format the step names
    spr_string <- paste0("%0", pad,"d")
    names(saved_complete) <- paste0("step_",sprintf(spr_string,seq(0,maxiter)))
    
    out_inp <<- vector(mode="list",length=length(maxiter))
    out_res <<- vector(mode="list",length=length(maxiter))
    out_bal <<- vector(mode="list",length=length(maxiter))
    
    timing <- matrix(NA,ncol=5,nrow=maxiter)
    msg("Loading phreeqc db... ")
    phreeqc::phrLoadDatabaseString(db)
    msg("database loaded.")
    if (procs > 1) {
        parallel::clusterCall(cl=ThisRunCluster, phreeqc::phrLoadDatabaseString, db)
        if (surrogate) {
            parallel::clusterExport(cl=ThisRunCluster, varlist=c("model","surrogate.FUN"), envir = environment())
            msg("All workers are setup with the surrogate model.")
        }
        
    }
    if (missing(init)) {
        msg("missing initial chemical state")
        if (is.null(initsim))
            initsim <- c(first,base)

        msg("running initsim PHREEQC script assuming EQUILIBRIUM and correct PUNCH specifications...")
        tmpfirstrun <- RunPQC(initsim)
        msg("assuming homogeneous medium")
        state_C <- matrix(rep(tmpfirstrun,n), byrow=TRUE, ncol=length(prop))
        colnames(state_C) <- colnames(tmpfirstrun)
    } else {
        msg("given initial state; assuming correct colnames")
        state_C <- init
    }

    attr(state_C,"immobile") <- immobile

    ## we save with pe/pH
    index_saved <- 1
    saved_complete[[index_saved]] <- list(C=Act2pH(state_C))
    index_saved <- index_saved + 1

    ## transform the pH/pe back into activities for Advection
    msg("first step, Advection()...")
    state_T <- AdvectionPQC(pH2Act(state_C), inflow=bound, dx=dx, dt=dt)
    Tr <- Ts <- 0

    ## treat special H+/pH, e-/pe cases
    state_T <- Act2pH(state_T)
  
    if (reduce)
        Tr <- system.time( reduced <- ReduceState(state_T) )[3]
    else
        reduced <- state_T
    
    msg("first step, chemistry ...")
    Ts <- system.time(inplist <- splitMultiKin(data=reduced, procs=procs, base=base, first=first, ann=ann, 
                                               prop=prop, minerals=immobile, kin=kin, dt=dt, nmax=maxsim))[3]
    
    Tm <- system.time(tmp <- RunPQC(inplist, procs=procs))[3]
    res_C <- tmp[, prop]
    attr(res_C,"immobile") <- immobile
    
    msg("iteration 0, CPU-time = ",round(Tm,3),"[s]")

    ## recompose after the reduction
    if (reduce)
        Tr <- Tr + system.time(state_C <- RecomposeState(res_C, reduced))[3]
    else {
        state_C <- res_C
        immobile_col_names <- setup$prop[setup$immobile]
        immobile_inds <- match( immobile_col_names, colnames(state_C) )
        attr(state_C,"immobile") <- immobile_inds
    }
    
    saved_complete[[index_saved]] <- list(T=Act2pH(state_T), C=Act2pH(state_C))
    index_saved <- index_saved + 1
    out_inp[[1]] <<- inplist
    out_res[[1]] <<- res_C
    
    if (ebreak) {
        msg("Early break invoked, bye.")
        return(saved_complete[[1]])
    }
    time <- dt
    iter <- 1

    timing[iter,] <- c(dim(reduced)[1],procs,Tm, Tr, Ts)
    
    msg("going through iterations now!")
    
    on.exit({
        msg("MAIN LOOP interrupted during iteration",iter,"!!")
        msg("returning last calculated chemistry.")
        print(traceback())
        msg(" Bye.")
        attr(state_C,"timing") <- timing
        saved_complete$currentT <- Act2pH(state_T)
        saved_complete$current <- Act2pH(state_C)
        return(saved_complete)
    })
    
    ## start of the iterations
    while (iter<maxiter) {
        ## iterations
        # TODO: It seems to me that iteration numberins is off by one, 
        # so that the first run on Rphree is actually one and the next ones end
        time <- time + dt
        iter <- iter + 1
        #cat("ENTERING ADV","\n")

        ## transform the pH/pe back into activities
        state_T <- AdvectionPQC(pH2Act(state_C), inflow=bound, dx=dx, dt=dt)

        ## reduction of the problem
        if(reduce)
            Tr <- system.time(reduced <- ReduceState(state_T))[3]
        else
            reduced <- state_T

        ## transform the activity back into pH/pe
        reduced <- Act2pH(reduced)

        ## this is the place where we switch the simulator to the surrogate model
        ## for final test for the geochemical surrogate model!
        
        if (surrogate) { ## use a surrogate model created by the Surrogate playground script
            Ts <- 0.0        ## no split and distribute for this guy
            
            Tm <- system.time( tmpsur <- Selected.Surrogate(reduced, model=model) )[3]
            ## print(head(tmpsur))
            bal <- CheckBalance(baleq,reduced,tmpsur)
            out_bal[[iter]] <<- bal
            totbal <- apply(bal, 1, mae)
            
            ## cat(":: Surrogate Balance Summary:\n")
            ## print(summary(totbal))
            ## tol2 <- unname(quantile(totbal, .85))
            ## cat(" :RT: Balance Checked\n")
            
            nonok <- which(totbal>tol)
            if (length(nonok)>0){
                msg("found", length(nonok), "/", nrow(reduced), " offending simulations",
                    ifelse(call_pqc, ", calling PHREEQC"," "))
                if (call_pqc) {
                    reducednonok <- reduced[nonok, ]
                    Ts <- Ts + system.time(inplist <- splitMultiKin(data=reducednonok, procs=procs, base=base, first=first, ann=ann,
                                                                    prop=prop, minerals=immobile, kin=kin, dt=dt, nmax=maxsim))[3]
                    
                    out_inp[[iter]] <<- inplist
                    if (length(nonok)==1) {
                        Tm <- Tm + system.time(tmp2 <- RunPQC(inplist, procs=1))[3]
                        pqcres_C <- tmp2[,prop]
                    } else {
                        Tm <- Tm + system.time(tmp2 <- RunPQC(input=inplist, procs=procs))[3]
                        pqcres_C <- tmp2[,prop]
                    }
                    
                    tmpsur[nonok, ] <- pqcres_C
                }
            } 
            res_C <- tmpsur
            attr(res_C,"immobile") <- immobile
        } else { ## no surrogate, the original implementation
            ## new input
            Ts <- system.time(inplist <- splitMultiKin(data=reduced, procs=procs, base=base, first=first, ann=ann,
                                                       prop=prop, minerals=immobile, kin=kin, dt=dt, nmax=maxsim))[3]
            out_inp[[iter]] <<- inplist
            Tm <- system.time(tmp3 <- RunPQC(inplist, procs=procs))[3]
            res_C <- tmp3[,prop]
            attr(res_C,"immobile") <- immobile
        }
        
        ## recompose after the reduction
        if (reduce)
            Tr <- Tr + system.time(state_C <- RecomposeState(res_C, reduced))[3]
        else {
            state_C <- res_C
            immobile_col_names <- setup$prop[setup$immobile]
            immobile_inds <- match( immobile_col_names, colnames(state_C) )
            attr(state_C,"immobile") <- immobile_inds
        }

        ## transform the pH/pe back into activities
        state_C <- pH2Act(state_C)

        out_res[[iter]] <<- res_C
     
        timing[iter,] <- c( dim(reduced)[1], procs, Tm, Tr, Ts)
        msg("done iteration", iter, "/", maxiter," CPU-time ",round(Tm,3),"[s]")
        
        saved_complete[[index_saved]] <- list(T=Act2pH(state_T), C=Act2pH(state_C))
        index_saved <- index_saved + 1
        
    }
    on.exit()

    if (procs > 1) {
        stopCluster(ThisRunCluster)
        msg("Parallelization cluster stopped")
    }

    msg("total time of chemistry: ",round(sum(timing[,3]),3),"seconds")
    msg("total number of simulations: ",sum(timing[,1]))
    msg(" Bye.")
    
    attr(saved_complete,"timing") <- timing
    #cat("PASS iteration saving ")
    return(saved_complete)
}
