##  Functions for dealing with surrogate simulations

### Marco De Lucia, delucia@gfz-potsdam.de, 2009-2018
### Janis Jatnieks, janisj@gfz-potsdam.de, jatnieks@janis.es
### Time-stamp: "Last modified 2018-05-09 01:07:22 delucia"
### Licence: LGPL version 2.1

## resolve dependencies automatically
## put neccessary packages here, they will be checked, loaded and autoinstalled if neccessary
##' @export
start_up <- function(session_path=FALSE, ## to disable set to a non-string
                     list.of.packages = c(## infrastructure and data manipulation
                         "zoo","plyr","sqldf","dtplyr","foreach",
                         "data.table",
                         ## multi-method learning meta-packages
                         "DiceEval","caret", #,"caretEnsemble" 
                         ## actual learning method packages
                         "mda","gam","polspline","rpart","MASS",
                         "elasticnet","e1071","ipred",
                         "deepnet","kernlab","pls","fastICA","lars",
                         "monomvn","RSNNS","qrnn",
                         "party","quantregForest","arm","brnn",
                         "lattice","ggplot2","plotrix", ## viz prototyping 
                         ## testing
                         "cluster","stringr","gtools", #"ftsa", 
                         "amap"),
                     install=TRUE,
                     use_cores=4)
{
    gc()
    
    msg("loading doParallel")
    list.of.packages = c(list.of.packages,"doParallel")
    
    ## load session data first
    ## if (class(session_path)[1]=="character" & file.exists(session_path) ) { 
    ##     msg("Loading...",session_path)
    ##     load(session_path, .GlobalEnv)
    ##     msg(" OK")
    ## }
    ## loading the right data is important!
    ## check install and load neccessary packages
    msg("Checking for required packages...")

    new.packages <- list.of.packages[!(list.of.packages %in% installed.packages()[,"Package"])]
    
    if(length(new.packages)) {
        cat("Need to install these packages:",new.packages,"\n")
        if (install) {
            libpath <- readline(prompt="Please specify the absolute path where the packages will be installed (e.g., '~/Rdevel'): \n")
            cat(" Afterwards, remember to store this path somewhere for R to find it again, such as R_LIBS_USER='~/Rdevel' in ~/.Renviron\n")
            install.packages(new.packages, lib=libpath)
        }
    } else msg("Package checks OK!")
    
    loadsuccess <- lapply(list.of.packages, require, character.only=TRUE)
    
    msg("Registering parallelization with ", use_cores, "cores")
    registerDoParallel( use_cores ) ## set up parallelization

    ### MDL: some global definitions
    ## we use this to swich calls between Dice and caret or direct calls
    ## to others
    DiceMethods <<- c('Linear','Additive','MARS','MARS3',
                      'PolyMARS','PolyMARS3','StepLinear','PolyMARS_GCV1')
    
    ## I enumerate them here so that no large text blocks 
    ## would repeat in pre- and post- processing code
    specific_preprocessing_methods <<- c(
        "(0,1) scaled rolling deltas",                 ## 1
        "(-1,1) scaled rolling deltas",                ## 2, usually not worth it
        "z-scaled rolling deltas",                     ## 3
        "z-scaled",                                    ## 4
        "deltas on z-scaled",                          ## 5
        "deltas on (0,1) scaled",                      ## 6
        "deltas on (-1,1) scaled",                     ## 7, usually not worth it
        "(0,1) scaled",                                ## 8
        "(0,10) scaled",                               ## 9
        "(-10,10) scaled",                             ## 10
        "(0,10) scaled rolling deltas",                ## 11
        "(-10,10) scaled rolling deltas",              ## 12
        "(0,1) scaled substract",                      ## 13
        "(0,10) scaled substract",                     ## 14
        "(-10,10) scaled substract",                   ## 15
        "z-scaled substract",                          ## 16
        "no",                                          ## 17
        "pca",                                         ## 18
        "ica",                                         ## 19
        "BoxCox",                                      ## 20
        "YeoJohnson",                                  ## 21
        "expoTrans",                                   ## 22
        "range",                                       ## 23
        "scale",                                       ## 24
        "spatialSign",                                 ## 25
        "center and scale"                             ## 26
    )
    
    msg("Startup OK!")
}


##' @export
SelectColsByPredix <- function(table, prefix)
{
    subset(table, select=grepl(prefix,colnames(table)))
}

## supply a vector or list of data str names to remove from global env
## only remove those that exist to avoid tons of warnings; also assert
## that none of the specifed structures really are in env after rm
## call
##' @export
mrm <- function(v, verbose=FALSE)
{
    to_remove <- v[laply(v,.fun=function(i) exists(i,where=.GlobalEnv))]
    ## only do something if any of given datastrs are bound to the env
    if ( length(to_remove)>0 ) { 
        if ( verbose )
            cat("\nRemoving existing datastructures:",to_remove)
        rm( list = to_remove, envir=.GlobalEnv ) 
        not_exist <- v[ laply(v,.fun=function(i) !( exists(i,where=.GlobalEnv)) ) ]
        if ( identical( v, not_exist) ) {if (verbose) cat("...check clean!\n")
        } else if (verbose) {
            cat("...unable to remove:",v," from ",not_exist,"\n")
        }
    }
}

## construct formula string from all column names in dataframe/table
##' @export
cfa <- function(y, table) {
    paste(y,'~',striplast(paste( colnames(Cout),"+",collapse=" ")),collapse=" ")
}

## construct formula sting from all column names in dataframe/table
##' @export
cf <- function(i,table) {
    paste(colnames(table)[i],'~.',collect="") }

## extract named or numbered vector from data table, frame or matrix
##' @export
dt2v <- function(sd=fitOUT, name="Ca") { as.vector(as.matrix( subset(sd, select=name) ))  }

## round every col differently according to spec given in list with
## corresponding names
##' @export
roundbycol <- function(t,spec) {
    ## check if columns in spec are also in the table then do it
    round_names <- intersect(names(spec),colnames(t))
    out<-as.data.table(llply(round_names,.fun=function(colname){
        round(dt2v(t,colname),spec[[colname]])},.parallel=TRUE) )
    colnames(out)<-round_names
    return(out)
}

##' @export
smartround <- function(t, roundto=Inf ) {
    if (!roundto==Inf) {
        
        if ( class(roundto)=="list" ) { 
            out <- roundbycol(t,roundto) }
        else { 
                out <- round(t, roundto) }
    }
    else { out <- t }
    return(out)
}

## check head and tail of the surrogate output to response table
##' @export
CHT <- function(real_out=data.table(valOUT), s_outp=SRes) { 
    ## print( real_out )
    ## print( s_outp )
    ident <- identical(real_out,s_outp)
    cat("\nIDENTICAL OUTPUTS?:",ident,"\n")
    if (!ident) {
    }
}

### Parameter overlap determination
## we check for prediction errors, find their indices thus determining
## the parameter space regions where they reside
##' @export
paralap <- function(real_out=data.table(valOUT), s_outp=SRes, inp=data.table(valIN) ) { 
    ## 1. substract output values from predictions
    dif <<- outminus(SRes,data.table(real_out)) 
    ## 2. get the indexes where difference>0 for each output column 
    colerridx <<- llply(seq(ncol(dif)), .fun=function(i) {seq(nrow(dif))[dt2v(dif,i)>0]},.parallel=TRUE )
    names(colerridx)<<-colnames(real_out)
    ## all error indexes 
    allerridx <<- sort(unique(unlist(colerridx,use.names=F)))
    ## 3. collect actual values that are mispredicted at their idx locations
    allcolerrvals <<- llply(seq(ncol(dif)), .fun=function(i) { dt2v(SRes,i)[ colerridx[[i]] ]},.parallel=T )
    colerrvals    <<- llply(allcolerrvals,.fun=function(vals) {sort(unique(vals))} )
    colerrfreq    <<- llply(seq(ncol(dif)), .fun=function(i) { as.numeric(ftable(allcolerrvals[i])) })
    names(colerrvals)   <<-colnames(real_out)
    names(allcolerrvals)<<-colnames(real_out)
    names(colerrfreq)   <<-colnames(real_out)
    ## uniqerrvals<<-llply(colerrvals, .fun=function(vals) { unique(vals) },.parallel=T )
    ## 4. how many real values are the same as the error values
    overlapcount <<- llply(seq(ncol(dif)), .fun=function(i) { 
        ## for each error value find where its duplicates are
        as.numeric( 
            llply(colerrvals[[i]],.fun=function(errval) {
                ## how many real values overlap the with the mispredicted value
                sum( dt2v(real_out,i)==errval )
            }))
    },.parallel=T )
    ## 5. collect input parameters for the erroenous predictions
    errparas <<- llply(seq(ncol(dif)), .fun=function(i) { inp[ colerridx[[i]], ]},.parallel=TRUE )
    return(overlapcount)
}

## unique values in column, to retrieve samples from design table
## takes data.table or data.frame
##' @export
uvc <- function(t) { 
    out <- llply( colnames(t), .fun=function(col) { unique( dt2v(t,col) ) }, .parallel=TRUE )
    names(out) <- colnames(t)
    return(out)
}

##' @export
uvcm <- function(t){ lapply(uvc(t),FUN=function(col) min(col) ) }

## unique column value dataframe
##' @export
uvt <- function( inT, outT, fun=coluniratio ) { 
    common_names <<- base::intersect( colnames(inT), colnames(outT))
    data.frame(OUT = fun( subset(outT,select=common_names) ),
               IN  = fun( subset(inT, select=common_names) ),
               row.names=common_names) }

# ratio of unique values among all 
##' @export
coluniratio <- function(t, roundto=Inf) {
    round(coluniquevals(t,roundto=roundto)/nrow(t),4)
}

##' @export
coluniquevals <- function(t, roundto=Inf) {
    laply(colnames(t), .fun=function(col) {
        if (!roundto==Inf) {colvals <- round ( dt2v(t,col), roundto )}
        else {colvals <- dt2v(t,col)}
        uvals   <- length( unique(colvals) )
        return(uvals)
    })
}

##' @export
allcoluniquevals <- function(t, roundto=8, para=TRUE) {
    auvals <- llply(colnames(t), .fun=function(col) {
        colvals <- round ( dt2v(t,col), roundto )
        uvals   <- sort(unique(colvals))
        return(uvals)
    },.parallel=para)
    names(auvals) <- colnames(t)
    return(auvals)
}

## how many values are shared between training and validation sets
##' @export
sharedvalues <- function(t1,t2, roundto=Inf, para=TRUE) {
    if (! (length(intersect( colnames(t1),colnames(t2)))==length(colnames(t1))  ))  {
        stop("ERROR: The given tables have different column names!")}
    svals <- llply(colnames(t1), .fun=function(col) {
        col1 <- dt2v( t1, col )
        col2 <- dt2v( t2, col )
        length( intersect(col1,col2) )/(nrow(t1) + nrow(t2))
    },.parallel=para)
    names(svals) <- colnames(t1)
    return(svals)
}

##' @export
SelectActiveColumns <- function(t) {
    ## return the data.frame/ table where columns have more than 1 value
    coluvals <- coluniquevals(t)
    ## cat("\nColumn unique val counts:\n",coluvals,"\n\n")
    subset(t, select=seq(ncol(t))[!(coluvals==1)] )
}

## return the data.frame/ table where columns have more than 1 value
##' @export
SelectMinActiveColumns <- function(t,mincount=5) {
    coluvals <- coluniquevals(t)
    ## cat("\nColumn unique val counts:\n",coluvals,"\n\n")
    subset(t, select=seq(ncol(t))[!(coluvals<mincount)] )
}

##' @export
nan.to.zero <- function(dfrm) { d2 <- dfrm
                                d2[is.nan(d2 <- dfrm)] <- 0
                                d2 
}

##' @export
na.to.zero <- function(dfrm) { d2 <- dfrm
                               d2[is.na(d2 <- dfrm)] <- 0
                               d2 
}

##' @export
to.zero <- function(dfrm,testfunc=is.na) { 
  d2 <- dfrm
  d2[testfunc(dfrm)] <- 0
  d2 
}

##' @export
colwise_RSS <- function( t1,t2 ) {
    cols <- union( colnames(t1), colnames(t2) ) # make sure we compare the same stuff
    data.frame( RSS=laply(cols,.fun=function(colname) { 
        sum( (subset(t1,select=colname)-subset(t2,select=colname))^2 )
    },.parallel=TRUE),row.names=cols )
}

##' @export
RSS <- function(v1,v2) { sum( (v1-v2)^2 ) }

## abs max error
##' @export
AME <- function(v1,v2) { max(abs(v1-v2)) }

## sum of absolute deviations
##' @export
SAD <- function(v1,v2) { sum( abs(v1-v2) ) }

## mean of absolute deviations
##' @export
MAD <- function(v1,v2) { mean( abs(v1-v2) ) }

##' @export
namewithin <- function(s1,number,s2) { paste0(s1,number,s2,collect="") }

##' @export
striplast <- function(s) { striplastx(s,1) }

##' @export
striplastx <- function(s,n=1) { substr(s,1,nchar(s)-n) }

## random sample of idxes from supplied number of rows in design dable
## and percentage ratio of how much should be sampled
##' @export
rsi <- function(n,ratio) { sample( n, n*ratio ) }

## col-wise rolling deltas
##' @export
Deltas <- function(dfrm) {
    dfrm[2:nrow(dfrm),]-dfrm[1:nrow(dfrm)-1,]
}

## Obtain min max and max-min range of data frame columns
##' @export
GetRanges <- function(dfrm) {
    savenames <- names(dfrm)
    ranges <- ldply(dfrm,.fun=function(col) { mm=range(as.numeric(col),na.rm=T,finite=T); c("min"=mm[1], "max"=mm[2] ) },
                    .parallel=TRUE,.id=NULL )
    rownames(ranges) <- savenames
    return(ranges)
}

## DATA EXTRACTION AND RESHAPING FOR MARCO'S RT EXAMPLE - Note, the
## res_full from Marco's example script must be loaded reshaping the
## data makes it easier to work with
##' @export
TByElem <- function(list_of_elems, nested_index=1,
                    attach_transport_timesteps=FALSE,
                    attach_element_numbers=FALSE)
{ 
    
    if (nested_index>0) { 
        elems <- nrow(list_of_elems[[1]][[nested_index]]) # get the number of elems in nested str
        out <- data.table( ldply( seq( length(list_of_elems)-1),.fun=function(i) {
            list_of_elems[[i]][[nested_index]]}
            )) 
    } else { 
        elems <- nrow(list_of_elems[[1]])
        out <- data.table( ldply(list_of_elems))
    }
  
    if (attach_transport_timesteps) {
        if (nested_index==1) {
            timesteps <- length(list_of_elems)-1 
            Ts <- rep( seq(1,timesteps,1), elems )
            out[,timestep:=Ts]
        }
    }
  
    if (attach_element_numbers) {
        elems <- sort(Ts)
        out[,ELEM:=elems]
        keycols=c("ELEM","timestep")
        keycols=c("timestep")
        setcolorder(out, union( keycols, setdiff(names(out),keycols )))
        if (nested_index==1) {out[,timestep:=Ts]}
    } 
    return(out)
}

## takes a table and returns it scaled to a new range, while allowing
## to define custom starting range for each col this is specific to
## the element-wise surrogate approach because the output is re-used
## as input in other cells they need the same normalization across
## both input and output
##' @export
rescaling <- function(t, old_ranges_table=FALSE, target_min=0,
                      target_max=1, target_ranges_table=FALSE)
{ 
  #TODO: (Fin[1,]-attr(Tin,"scaled:center"))/attr(Tin,"scaled:scale") #MAKE THIS INLINE
  
  ## if target ranges are not supplied for each column, then create
  ## this table from defaul 0 and 1
    if (target_ranges_table==FALSE) { 
        target_ranges <<- data.frame( 
            row.names = rownames(originalIOranges),
            min = rep(target_min,length(rownames(originalIOranges)) ),
            max = rep(target_max,length(rownames(originalIOranges)) )
        )
    }
    ## if ( target_min==NULL || target_max==NULL) stop("error rescaling:: No target range supplied!")
    if ( target_min > target_max)
        stop(paste("error rescaling:: min exceeds max",target_min,target_max))
    if ( !(class(old_ranges_table)== "data.frame") )  {
        old_ranges_table <- GetRanges( t )
        cat("Rescaling calculated initial ranges for columns as no custom ranges supplied!\n")
    }
    sapply(colnames(t),FUN=function(colname) { 
        custom_scaler( dt2v(t,colname),
                      old_ranges_table[colname,"min"],old_ranges_table[colname,"max"],
                      target_ranges[colname,"min"],target_ranges[colname,"max"])
    })
}

## substract columns, but only those dt2v(Deltas(Oout),1) present in
## the output table
##' @export
outminus <- function(inp, outp) {
    ## substract the the output columns from 
    vect <<- llply( colnames(outp),.fun=function(col) { dt2v(outp,col)-dt2v(inp,col) },.parallel=TRUE )
    names(vect) <<- colnames(outp)
    as.data.table(vect)
}


##' @export
colwise_rescale <- function(t, minval=0, maxval=1) {
    data.frame(llply(t,.fun=function(colu) { 
        rescale(colu,c(0,1)) }, .parallel=TRUE ))
}

##' @export
custom_scaler <- function(num_vect,
                          from_min,from_max,
                          to_min,to_max) {
                              ((to_max-to_min)*(num_vect-from_min))/(from_max-from_min)+to_min
}

## Mean absolte scaled error,
## http://robjhyndman.com/papers/foresight.pdf
## http://stackoverflow.com/questions/11092536/forecast-accuracy-no-mase-with-two-vectors-as-arguments
##' @export
MASE <- function(f,y) {
    ## f = vector with forecasts, y = vector with actuals
    if(length(f)!=length(y))
        stop("Vector length is not equal") 

    n <- length(f)
    res <- mean(abs((y - f) / ((1/(n-1)) * sum(abs(y[2:n]-y[1:n-1])))))
    return(res)
}

##' @export
FastClust <- function(dfrm, clustnum=7, g=TRUE, link_type="complete") {
    require(amap) # provides parallel distance calc, default dist stalls with small data

    hc <- hcluster(dfrm, method="euclidean",link=link_type,
                   nbproc=safe_get_cores(), diag=TRUE, upper=FALSE)
    if (g)
        plot(hc)
    
    cms <- data.frame( membership=cutree(hc,clustnum) )
    clusterIDs <- unique( cms$membership )
    llply( clusterIDs, .fun=function(clnum) row.names(subset(cms, subset=(membership==clnum) ) ) )
}


##' @export
safe_get_cores <- function() {
    ## register parallelization backends on anyhing but win, due to missing fork()
    require(doParallel)
    ## detect using 
    report_cores <- detectCores() 
    return( report_cores )
}

##' @export
BestModelCluster <- function(dt,by="C",criteria="MASE",clusters=10) {
    as.numeric( FastClust(subset(dt[(output == by)],select=criteria), clustnum=clusters)[[1]] )
}

## non-propogating simulation error controller takes all input
## matrices from RT run from simulator and uses SelectedSurrogate on
## each of them seperately to predict the simulator outputs at each
## step. WARNING: this does not imply input-output re-use as
## neccessary for full reactive transport emulation. This is simply
## aimed at debugging!
##' @export
NPSC <- function(simulation_data = res_sim,
                 inputs_name = "T",
                 outputs_name= "C",
                 in_round  = Inf, # re-round input that goes into the predictor
                 out_round = Inf, # round the output that comes out
                 sim_round = Inf, # round the simulator output used for comparison
                 show_head = 0, # show this much in head and tail of the error matrix
                 para=TRUE) {
    r <- llply(seq(length(res_sim)-1),function(i) {
        cat("\n",i," ")
        
        r <- outminus(smartround(simulation_data[[i]][[outputs_name]],sim_round),
                      smartround(
                          SelectedSurrogate(
                              smartround( simulation_data[[i]][[inputs_name]],in_round)),out_round) )
        if (show_head>0) { 
            cat("\n")
            print(head(r,show_head)) 
            print(tail(r,show_head)) 
        }
        return(r)
    },.parallel=para)

    return(r)
}

## test surrogate error drift when called repeated on own IO
##' @export
SDrift <- function( start_state = res_sim[[100]][["T"]], r = 50, head=10 ) {
    lastsur <<- SelectedSurrogate(start_state)
    lapply(seq(r),FUN=function(i){
        lastsur <<- head( SelectedSurrogate(lastsur) )
    })
}

##' @export
Tminus <- function(simulator_output=res_sim, 
                   surrogate_output=res_sur, 
                   nested_name="C", 
                   show_head=0 ) {
    if ( length( simulator_output)==length(surrogate_output) ) {
        r <- llply( seq(length(simulator_output)-1),
                   function(i) {
                       cat("\n",i," ")
                       r <- outminus( simulator_output[[i]][[nested_name]], surrogate_output[[i]][[nested_name]] )
                       if (show_head>0) {
                           cat("\n")
                           print( head(r,show_head) ) 
                           print( tail(r,show_head) ) 
                       }
                       return(r)
                   })
    } else {
        cat("FAIL: Simulator and surrogate output length differ!")
    }
}

##' @export
CompareTimesteps <- function(simdata     = res_sim,
                             surdata     = res_sur,
                             timestep    = 100,
                             nested_name = "C",
                             namepref    = paste0(GetModelNames()[[1]]," (",current_method,") ",current_preproc))
{
    tryCatch({
        simres <- simdata[[timestep]][[nested_name]]
        surres <- surdata[[timestep]][[nested_name]]
        
        writecols <- intersect( colnames(simres),colnames(surres) )
        simnewnames <- paste0(writecols, "_simulated")
        surnewnames <- paste0(writecols, "_predicted")
        writedata <- cbind( simres[,writecols], surres[,writecols] )
        
        colnames(writedata) <- c(simnewnames,surnewnames)
        if (!exists("ens_batch_name")) {
            ens_batch_name <<-paste0("SimSur_data/",namepref,"_step_",timestep,"_compare.csv")
        }
        if (!exists("ens_perftop_name")) {
            ens_perftop_name <<-paste0("SimSur_data/",namepref,"_step_",timestep,"_perftop.csv")
        }
        write.csv(writedata,ens_batch_name)
        write.csv(perftop,ens_perftop_name)
        mrm(c("ens_batch_name","ens_perftop_name"))
    },# end of tryCatch
    warning = function(cond) {
        message(cond)
    },
    error = function(cond) {
        cat("\nError when comparing timesteps!\n")
        ## print(head(writedata))
        message(cond)
    })
}

##' @export
ShowTopProfiler <- function(profiler_output, top=100, show=T) {
    profiler_summary <- summaryRprof( filename = profiler_output )
    pr <- head(profiler_summary[[2]],top)
    if (show) {
        cat('Top functions by\n')
        print( pr )
    }
    return(pr)
}

## exclude column using subset and set difference from dfrm/DT/matrix
##' @export
exc <- function(table_like, exc_cols) {
    exclude_cols_actually_present <- exc_cols[exc_cols %in% colnames(table_like)]
    subset(table_like, select <- setdiff(colnames(table_like),exclude_cols_actually_present) )
}

##' @export
SparseChangeCluster <- function(out_dfrm) {
    ## 1. get differences by changing 
    Deltas(out_dfrm)
    ## 2. estiamate initial number of clusters for this output table (can be 1 or more outputs)
    
    ## 3. get input value ranges for the clusters
    
    ## 4. get output value ranges for the clusters
    
    ## return a structure that has input and output clusters
}

##' @export
RangeTableCreator <-function(original_table,extra_table,col=1) {
    ## iterate over member list, take members subset original table by
    ## members, get range write range in return list
    col_member_list <<- FastClust( extra_table[[col]],3)
    col_ranges_list <<- llply(col_member_list,.fun=function(inds) {range(original_table[[col]][as.numeric(inds)]) })
    return(col_ranges_list)
}

## find specific types of columns in data.table/frame
##' @export
coltypes <- function(intable) {
    sapply(intable, FUN=class)
}

##' @export
seekcoltypename <- function(intable, seektype) {
    names(which( coltypes(intable)==seektype))
}

##' @export
seekcoltypeidx <- function(intable, seektype) {
    array(which( coltypes(intable)==seektype))
}

##' @export
f2i <- function(intable,seektype) {
    factcols <- seekcoltypename( intable, seektype )
    rtable <- exc( intable, factcols)
    if (length(factcols)>0) {
        enumcols <- as.data.table(sapply(factcols, FUN=function(fcol) {
            datavals <- dt2v(intable,fcol)
            fmap <- unique(datavals)
            names(fmap) <-seq(length(fmap))
            match(datavals,fmap)
        }))
        out<-cbind(rtable,enumcols)
    }
    else out<-intable
    return(out)
}

##' @export
exc_type <- function(intable, exclude_type="factor") {
    exc(intable, seekcoltypename(intable,exclude_type) )
}

##' @export
take_top_pct_cols <- function(indata, pct=0.33 ) {
    sqrd <- order(indata^2)
    alen <- length(sqrd)
    topx <- sqrd[1:round(alen*pct,0)]
    return(rownames(indata)[topx])
}

##' @export
mda_mars_importance <- function(mda_mars_model) {
    fsum <- colSums(abs(mda_mars_model[["model"]][["factor"]] ))
    names(fsum)[fsum > 0]
}

##' @export
all_mda_mars_impvars <- function( modellist )
{
    sapply( modellist, FUN=mda_mars_importance )
}

##' @export
CheckColsNumValues <- function(dfrmlike)
{
    col_uniq_values <- coluniquevals( dfrmlike )
    names(col_uniq_values) <- colnames(dfrmlike)
    return(col_uniq_values)
}

##' @export
Cols2Fact <- function(dfrmlike, thresh=100, para=TRUE)
{
    cuv <- CheckColsNumValues(dfrmlike)
    cuv <- cuv[cuv<=thresh]
    if (length(names(cuv))>0) {
        cat("Columns with low number of unique values:\n")
        print(cuv)
        fcs <- as.data.table( llply( names(cuv), .fun = function(col) { factor(dt2v(dfrmlike,col)) },.parallel=para ) )
        colnames(fcs) <- names(cuv)
        out <- cbind( subset(dfrmlike,select=setdiff(colnames(dfrmlike),names(cuv)) ), fcs )
    } else {
        out <- dfrmlike
    }
    return(out)
}


########################################################################
#### from here functions dealing with machine learning infrastructure,
#### dice and caret

## caret needs an advanced parameter tuning grid as data.frame for the
## models that depend heavily on funky parametrization for their
## effectiveness multiTuneGrids contain presumably sensible designs of
## such tuning grids singularTuneGrids take the most "epic" tune
## configuration and use it when tuning is disabled on caret train
## call, this means just direct fitting, to supplied data, without
## using the full tunegrid

## TODO: check traincontrol function for automatic fraction and
## resampling. MAYBE not everything is passed in

##' @export
createTuneGrids <- function() {
    cat('\nCreating tuning grids for caret models...\n\n')
    multiTuneGrids  <<- 
        list(
            "bayesglm"      = NULL,
            "rlm"           = NULL, ## tends to perform well and fast sometimes
            ## "gaussprLinear" = NULL ## consumes too much time / resources / dies
            "glm"           = NULL, 
            ##"nnls"          = NULL, ## the predict method is broken for this one, try after update to caret
            "lmStepAIC"     = NULL, ## decent for simpler cases, fast, but fails often too
            "glmStepAIC"    = NULL, ## decent for simpler cases, fast, but fails often too
            "leapBackward"  = list(data.frame ( nvmax = seq( 1, ncol(Fin)-1, 1 ) ) ),
            "leapForward"   = list(data.frame ( nvmax = seq( 1, ncol(Fin)-1, 1 ) ) ),
            "leapSeq"       = list(data.frame ( nvmax = seq( 1, ncol(Fin)-1, 1 ) ) ),
            "treebag"       = NULL,
            "lda"           = NULL,
            "ppr"           = list(data.frame (    nterms=seq(1,round(ncol(Fin)*.8,0),2))),  ## fast, often good
            ## qrnn very slow in training, decently fast and accurate in prediction, but no better than brnn and others good ones
            "qrnn"          = list(data.frame (    n.hidden=ncol(Fin)*2,penalty=2,bag=F ) ), 
            ## very slow in prediction, bad results
            ##"mlp"           = list(data.frame (    size=4 ) ) ## very slow and bad predictions
            ## these rbf guys are slow and give catastrophically horrifying results, maybe their parameters are completely off...
            ##"rbf"           = list(data.frame(size=5 ) ) ## very slow and bad predictions
            ##"rbfDDA"        = list(data.frame(negativeThreshold=0.1 ) ) ## really slow, mostly fails
            ##"Boruta"        = list(data.frame(mtry=ncol(Fin) ) ) ## slow, bad predictions
            "qrf"            = list(data.frame(mtry=7)), ## kind of slow in prediction, sometimes very good results
            "parRF"          = list(data.frame(mtry=7)), ## kind of slow in prediction, sometimes very good results
            "Rborist"        = list(data.frame(predFixed=ncol(Fin))), ## kind of slow in prediction, sometimes very good results
            ##"cforest"       = list(data.frame(mtry=ncol(Fin) ) ) ## resource hog, too slow in prediction, tolerable results, but not very good
            ## blackboost - not bad, but not very good, slowish also
            ##"blackboost"    = list(data.frame (    mstop=50, maxdepth=ncol(Fin) ) ) ## I also killed this one
            "kernelpls"     = list(data.frame(ncomp=seq(1,ncol(Fin)))), 
            ##"gaussprRadial" = list(data.frame(sigma=c(0.5,1) )) ## too slow
            "pcr"           = list(data.frame(ncomp   = seq( 1, ncol(Fin)-1))), ## perform well/decent
            "pls"           = list(data.frame(ncomp   = seq( 1, ncol(Fin)-1))),
            "simpls"        = list(data.frame(ncomp   = seq( 1, ncol(Fin)-1))),
            "enpls"         = list(data.frame(maxcomp = seq( 1, ncol(Fin),1))),
            "icr"           = list(data.frame(n.comp  = seq( 1, ncol(Fin),1))),
            ## lars(es) tends to perform well
            "lars"          = list(data.frame(fraction= c(seq(0.9,0.99,0.01),0.999))), 
            "lars2"         = list(data.frame(step    = seq(1,ncol(Fin),1))),
            "lasso"         = list(data.frame(fraction=c(seq(0.95,0.99,0.01),0.999))),
            "blassoAveraged"= NULL,
            "BstLm"         = list(data.frame(mstop=seq(ncol(Fin)),nu=0.5)), ## takes too long
            "enet"          = list( expand.grid(fraction=rev(seq(0.01,0.5,0.1)),
                                                lambda=seq(0.1,0.5,0.1))), ## bad predictions
            ## brnn good x-validation results, slow in training phase
            "brnn"          = list(data.frame(neurons=ncol(Fin)*2)), 
            ## drnn fast in training, weird tuning:
            "dnn"           = list(data.frame(layer1=100, layer2=100, layer3=1000,
                                              hidden_dropout=0.1,
                                              visible_dropout=0.1)),
            "elm"           = list(expand.grid(nhid=c(50,100),  ## fast in training, weird tuning
                                               actfun=c("purelin","poslin","tribas","tansig","satlins","hardlim",
                                                        "radbas","sin","sig"))),
            "bagEarth"      = list(data.frame(nprune=seq(4,10,1), ## slowish and weak
                                              degree=seq(4,10,1))),
            "bagEarthGCV"   = list(data.frame(degree=seq(1,ncol(Fin),1))),
            ## "nodeHarvest"   = list( expand.grid(    maxinter=rev(seq(1,ncol(Fin),1)),mode="mean") ) ## resource hog, slow with weak results

            ## this guy will need more understanding on tuning, maybe useful, but slow and much work
            "deepboost"     = list(data.frame (num_iter= 100,
                                               tree_depth=ncol(Fin),
                                               beta=0,
                                               lambda=0.00125,
                                               loss_type="e")),
           "logreg"    	 = list(data.frame (treesize =100, ntrees=100)), ## something about predicting binary stuff
           "ANFIS"    	 = list(data.frame (num.labels=5, max.iter=10)), ## way too slow...probably
           "gbm"         = list( expand.grid(n.trees = 1000, 
                                             ## n.cores = 8,
                                             interaction.depth = seq(2,5,1), 
                                             shrinkage = 0.0001, ## important to tune well
                                             n.minobsinnode = rev(seq(6,20,2)))),
           "xgbTree"        = list( expand.grid(nrounds  = c(700,1500,2000),
                                                max_depth = seq(2,30,2),
                                                eta  = c(0.01,0.1,0.15),
                                                gamma = c(0.01,0.05,0.1,0.2),
                                                colsample_bytree = 0.95,
                                                min_child_weight = c(5,10,100,25),
                                                subsample  = 0.95)),
           "xgbLinear"      = list( expand.grid(nrounds  = c(20,50,80,120,150,180,200,250,300,5000),
                                                alpha = 0,
                                                eta  = c(0.01,0.03,0.05,0.1,0.15),
                                                lambda = 0 )), ## performs well on Kaggle stuff, could be faster in running, but not bad
           ## "ctree"          = list(data.frame(mincriterion=0.99 )), ## does not perform well and is a resource hog
           ## "ctree2"         = list(data.frame(maxdepth=seq(2, ncol(Fin)*3, 1))), ## does not perform well
           ## our problem is unsuitable for widekernel
           "widekernelpls"  = list(data.frame(ncomp=seq(2,ncol(Fin)))), 
           ## "rvmLinear"     = NULL ## way way too slow
           ## cubist is really slow like 100x than others with this tunegrid
           "cubist"         = list(data.frame( neighbors=seq(3, 9, 1),committees=seq(3,9,1) )  ),
           ## this model seems to perform badly
           "svmRadialCost"  = list(data.frame(      C=c(  0.1, 0.001, 0.0001 ) ) ), 
           "ridge"          = list(data.frame( lambda=c(0,0.1, 0.001, 0.00001) ) ),
           "foba"           = list( expand.grid( k=seq(1, ncol(Fin) ), lambda=c(0.05)) ),
           "bridge"         = NULL,
           "rpart1SE"       = NULL,
           "rpart2"         = list( expand.grid( maxdepth = seq(1, 30, 1) ),
                                   list( cp=0.00001, minbucket=2,minsplit=2) )
           ## "rpart"         = list(data.frame( cp=c(0.0001,0.00001),  ## tune grid with officially tunable parameters
           ##                   list( minsplit=2, minbucket=2)        ))   ## other parameters that are given to train through passthrough
           ##                   minsplit=2, minbucket=2        ))   ## other parameters that are given to train through passthrough
        )

    ## for this we take the most extreme parameter combinations from the multi tune grid
    singularTuneGrids  <<- lapply( seq(1, length(multiTuneGrids), 1), FUN = function(i) {
        tune_content <- multiTuneGrids[[i]][[1]] ## take the tuneGrid data.frame
        if (class(tune_content)=="NULL") {
            result = NULL
        } else if (class(tune_content)=="data.frame") { 
            lastrow <- nrow(tune_content)
            result <- tune_content[lastrow,,drop=FALSE]
        }   

        if ( length(multiTuneGrids[[i]]) == 1) {
            result <- list( result )
        }

        if ( length(multiTuneGrids[[i]]) == 2) {
            result <- list( result, multiTuneGrids[[i]][[2]] )
        }
        return( result )
    })
    names(singularTuneGrids) <<- names(multiTuneGrids)
}

##' @export
FitWithDice <- function( fit_in, fit_out, m="PolyMARS_GCV1" ) {
  if (m=="MARS3"||m=="PolyMARS3") {
    m = striplast(m)
    model = modelFit( fit_in, fit_out, type=m, degree=ncol(fit_in), penalty=ncol(fit_in), thresh=0.0001, gcv=1, trace=T)
  }
  else if (m=="PolyMARS_GCV1") { 
    m = striplastx(m,5)
    model = modelFit( fit_in, fit_out, type=m, gcv=1) 
  }
  else if (m=="PolyMARS") { 
    model = modelFit( fit_in, fit_out, type=m, gcv=4) 
  }
  else if (m=="MARS") { 
    model = modelFit( fit_in, fit_out, type=m, degree=2) 
  }
  else if (m=="Linear") { 
    model = modelFit( fit_in, fit_out, type=m, degree=2) 
  }
  else { 
    model = modelFit( fit_in, fit_out, type=m ) 
  }
  return(model)
}

## learn the surrogate using supplied methods
##' @export
FitSurrogates <- function(## fitIN, fitOUT,
                          ## indices,
                          method = "MARS", ## assume all are MARS if not given
                          preprocessing_meth_inds = c(),
                          fitting_config=c(),
                          ## dot not fit the variable to self, if names match
                          exclude_self = FALSE,
                          tune = FALSE,
                          para = FALSE)
    ## there are two ways that this is called - in batch screening mode, when the model type is 
    ## 1. always the same, due to how iterators in main are built (the same method for all)
    ## 2. from refit, when the best model types are selected from screening with their preprocessing attached
{ 
    response_vector_names <<- paste0("response_vector_", colnames(Fout),sep="")
    mrm( c("PPCout","PPC","model_list", response_vector_names))
    t <- system.time( model_list <<- llply( seq( ncol(Fout) ), .fun=function(i) {
        ## first get the training data samples
        ## if indices are a list then, for each variable we take different ones    
        ## if ( class(inds)=="list" ) {
        ##  idxes <- inds[[i]]
        ##  fitIN <- Fin[ idxes, ] ## access the i-th column-wise resampling index vector
        ##  fitOUT<-Fout[ idxes, ] ## access the i-th column-wise resampling index vector
        ## } else { #}
        
        ## if ( length(inds)==nrow(Fin) ) { ## if all the data is used for training then just do it
        ##	fitIN <- Fin
        ##	fitOUT<- Fout
        ##} else {
        fitIN <- Fin[ inds, ]
        fitOUT<-Fout[ inds, ]
        ##}
        
        fitIN <- SelectActiveColumns(fitIN)
        fitOUT <- SelectActiveColumns(fitOUT)
        
        ## fit a separate model of this method for each col
        ## select all neccessary data and settings for fitting output-wise surrogate model
        tryCatch({
            outname <- colnames(fitOUT)[i]
            cat( outname,' ') ## show which output 
            resp_name <- response_vector_names[i]
            assign( resp_name, dt2v(fitOUT,outname) )#, envir=.GlobalEnv )
            ## if preprocessing vector has anything specified, then check that it contains 
            ## a spec for every output and then run it through PreProc
            if ( length( fitting_config )>0 ) {
                method  <- as.character( fitting_config[outname,"method"] )
                SelectPreProc( match(fitting_config[outname,"preprocessing"],specific_preprocessing_methods), v=F)
                cat(" ",method,as.character(fitting_config[outname,"preprocessing"]), "\n" )
                fitIN <- as.matrix( PPC(fitIN) )
                if (exclude_self) { 
                    cat("\nexcluding",outname,"\n")
                    fitIN <- exc(fitIN,outname)
                    ## print(head(fitIN))
                }
                ## if output also has to be preprocessed then do it again!
                if (exists("PPCout")) { 
                    ## use assign to create different names for each output response 
                    assign( resp_name, PPCout(fitIN, dt2v(fitOUT,outname) ) )#, envir=.GlobalEnv )
                }} 
            
            ## OK start fitting the models depending on the supplied method descriptions
            ## if (method %in% DiceMethods) { model = FitWithDice(fitIN, get(resp_name,envir=.GlobalEnv), method)  }
            if (method %in% DiceMethods) { model = FitWithDice(fitIN, get(resp_name), method)  }
  
            ## direct calls to some packages
            ## call mars directly with custom parameters
            if (method == "mda_p01")     { model = mars(fitIN, get(resp_name),#,envir=.GlobalEnv),
                                                        thresh=0.0000001,
                                                        ## penalty=ncol(fitIN),
                                                        penalty=round( ncol(fitIN)*0.9,0),
                                                        degree =round( ncol(fitIN)*0.9,0) ) }
            
            ## call mars directly with custom parameters
            if (method == "polymars_p01"){ model = polymars(get(resp_name), fitIN,
                                                            additive=TRUE, gcv=0.01 ) }
            
            ## attempt directly tunable CART fit
            if (method == "rpart_anova_direct")  {
                model <- rpart(cf(i,fitIN), as.data.frame(fitIN), 
                               method="anova",control=rpart.control(minsplit=3,
                                                                    cp=0.00001,
                                                                    minbucket=2))
            }
    
            ## all other that have tuning grids are assumed to be
            ## caret methods and trained through the caret interface
            if (method %in% names(multiTuneGrids)) {
                ## this gets a bit tricky. we support arbitrary
                ## parameters not allowed by train tuneGrid to do this
                ## we build a custom call for train that passes also
                ## these extra named params to tune which then passes
                ## them thruogh to the actual model fit call in the
                ## correct way
                ## if (!tune) { train_call  <<- list( as.name("fitIN"), as.name("response_vector"), method=method,
                if (!tune) {

                    train_call  <- list( as.name("fitIN"), as.name(resp_name), method=method,
                                        ## just fit a simple model to the supplied data directly
                                        trControl = caret::trainControl(method='none',returnData=FALSE,
                                                                        trim=TRUE, allowParallel=FALSE, p=1.0),
                                        ## retrieve the grid for each method
                                        tuneGrid  = singularTuneGrids[[method]][[1]] )
                }
                if ( length(singularTuneGrids[[method]]) == 2) { 
                    train_call <- c(train_call, singularTuneGrids[[method]][[2]] )
                }
                
                ## if ( tune ) { train_call  <<- list( as.name(fitIN), as.name("response_vector"), method=method,
                if ( tune ) {
                    train_call  <- list( as.name("fitIN"), as.name(resp_name), method=method,
                                        ## just fit a simple model to the supplied data directly
                                        trControl = caret::trainControl(method='boot', returnData=FALSE,
                                                                        trim=TRUE, search="grid", allowParallel=FALSE,
                                                                        number=2, verboseIter = TRUE),
                                        ## retrieve the grid for each method
                                        tuneGrid  = multiTuneGrids[[method]][[1]] )
                    
                    if ( length(multiTuneGrids[[method]]) == 2)  
                        train_call <- c(train_call, multiTuneGrids[[method]][[2]] ) 
                }
                
                ## TODO: FIX NEW API
                ## cat(paste("calling caret::train with call: ", str(train_call),"\n"))
                model <- do.call( caret::train, train_call )
            }
            
        },## end of tryCatch
        warning = function(cond) {
            message(cond)
        },
        error = function(cond) {
            cat("\nFailed @ fitting!\n")
            message(cond)
        }#, finally = {}
        )
        if (!exists("model")) {
            model="failed @ fitting"
        }
        return(model)
    },.parallel=para) )[[3]]
    
    cat('|',t,'s for all.\n')
    mrm( response_vector_names )
    names(model_list) <<- colnames(Fout)
}

## take a model list and check what they actually are
##' @export
GetModelNames <- function(models=model_list) {
    method_realname <- "undetermined"
    model_names <- llply( seq(length(models)),.fun=function(i) {
        if ("type" %in% names(models[[i]])) {  ## a Dice model description
            method_realname <- models[[i]][["type"]]
                                        #model_names <- c(model_names,method_realname)
        }
        else if ("modelInfo" %in% names(models[[i]]) ) { ## caret model info
            method_realname <- models[[i]][["modelInfo"]][["label"]] 
        }
        else { method_realname <- class(models[[i]]) 
            if (method_realname=="character") {
                method_realname=models[[i]]}
        }
        return(method_realname)
    })
    return(as.character(model_names))
}

##' @export
RunModelList <- function(inp, ## TODO: FIX INPUT PASSING AND VALIDATION SAMPLE CHECKING!
                         output_names, models = model_list,  ## method="MARS3",
                         custom_surrogate_inds = seq(length(model_list)),
                         inline_preprocessors = c(),
                         allow_negative_cols = c("e","pe","Pressure"),
                         exclude_self = FALSE,                         
                         para = FALSE ## usually the overhead is not worth it
                         )
{
    mrm(c("SRes"))
    
    ## output_names <- colnames(Fout)
    
    ## if only one name is supplied we assume they're all the same
    ## type of model

    ## if ( length(method)==1 && !(length(models)==1) ) {
    ## method=rep(method,length(models)) }
  
    ## those models that are of type character are failed fits
    successful_fits <<- seq(length(model_list))[!laply( model_list,.fun=class )=="character"]

    ## we also check if any where excluded by the user and pass only indexes in both for running
    run_idxes <<- intersect(successful_fits,custom_surrogate_inds)
    ## cat("\nSuccessfult fits",successful_fits,"run_idxes",run_idxes,"model_list",length(model_list),"\n")
    cat('Running surrogates for ')
    last_surrogate_runtime <<- system.time(
        SRes <- llply( run_idxes, .fun=function(i) {
            tryCatch({
                ## if the model_list is called together with a specification of
                ## preprocessing methods, then we use each one of them for inputs
                ## before applying the selected model from model_list
                if ( length( inline_preprocessors )>0 ) {
                    preproc_method_name <- inline_preprocessors[[i]]
                    SelectPreProc( match(preproc_method_name,specific_preprocessing_methods), v=F)
                    ## cat(" ",method,as.character(fitting_config[outname,"preprocessing"]), "\n" )
                    
                    ## inp <- as.matrix( PPC(inp) )
                    inp <- as.matrix( PPC(inp) )
                    ## if output also has to be preprocessed then do it again!
                    ## if (exists("PPCout")) { 
                    ## use assign to create different names for each output response 
                    ##  assign( resp_name, PPCout(fitIN, dt2v(fitOUT,outname) ) )#, envir=.GlobalEnv )
                }                  
                cat( output_names[[i]],' ' )
        
                if (exclude_self) { 
                    model_name <- names(models)[[i]]
                    inp <- exc(inp,model_name)
                    cat('\n')
                    ## print(head(inp))
                }
                ## detect the type of surrogate model that we have
                ## "list" = Dice model, "train" == caret, anythong else - try default predict method:) 
                model_type <- class(models[[i]])
                if ( model_type=="list" ) { 
                    ## if (exists("validation_data")) { out = modelPredict( models[[i]], validation_data )  }
                                        #else { 
                    out <- modelPredict( models[[i]], inp )
                                        #) }
                } else {
                    ## some models use a generic predict method
                    ## if (exists("validation_data")) { out = predict( models[[i]], validation_data ) }
                    ## else {
                    out <- predict( models[[i]], as.data.frame(inp) )
                }
            },
            warning=function(cond) {
                cat("WARNING: ", message(cond))},
            error = function(cond) {
                cat(" ... failed @ predicting!",message(cond))
                out=rep(NA, nrow(inp))},
            finally=function(cond) {
                out=rep(NA, nrow(inp)) }
            )  ## end tryCatch  
            ## final checks on predictor output
            if ( !exists("out") ) { 
                cat("WARNING: No predictor output was produced, without error, returning NAs!")
                out <- rep(NA, nrow(inp))
            } ## set all negatvie values to zero, if not forbidden-concentrations cannot be negative!
            ## if ( !(output_names[[i]] %in% allow_negative_cols)) {
            ##  out[out<0] <- 0
                                        #}
            return(out)}, ## end of llply
            .parallel=para ) )[[3]] ## end of system.time

    SRes <- as.data.table(SRes)
    ## cat("SRes",str(SRes))
    ## cat("output_names",output_names)
    
    ## it is important to assing the names of the output columns that
    ## were actually selected to be predicted and were a model that
    ## did not fail was successful at running
    if (ncol(SRes)>0) {
        colnames(SRes) <- output_names[run_idxes]
    } else
        cat("Warning: no prediction model(s) in this ensamble!")

    model_names <<- GetModelNames()
    return(SRes)
}

## output-wise for each surrogate method
##' @export
RunSurrogates <- function(out_col_names,
                          models=model_list, method="MARS3",
                          show_results=FALSE) { 
    
    out_col_names <- colnames(Fout)
    cat("CalledRunSurrogates\n")
    ## we only select validation samples as fraction of training if a
    ## special validation_data is not set
    if ( class(inds)=="list" ) {
        inputs = FALSE
    } else {
        inputs = Fin[ -inds, ]
    }
    
    SRes <<- RunModelList( inputs, out_col_names, model_list)#, method) ## this is reused when replacing the simulator 
    cat('|',last_surrogate_runtime,'s\n')
    out <<- as.data.table( SRes )
    time_per_output = ( last_surrogate_runtime/length(models) )
    
    cat( method, current_preproc, last_surrogate_runtime, time_per_output )
  
    Sperf[(`method`==current_method & `preprocessing`==current_preproc ),`speed`:=time_per_output ]
    Sperf[(`method`==current_method & `preprocessing`==current_preproc ),`method_realname`:=model_names ]
  
    SRes <<- out
  
    if (show_results) { 
        cat("\nSurrogate example output and actual output:\n")
        print( out ) 
        print( Fout[ -inds, ] )
    }
  
    return ( out )
}

##' @export
write_evals <- function( hist_bins = 25, write_errors=TRUE, Surrogate_performance_table = Sperf )
{
    ## cols <- intersect(colnames(valOUT), colnames(SRes))

    ## we do a pass of all columns that needed to be predicted and
    ## then check if everything is OK with the data
    l_ply( colnames(Fout), .fun = function(outputname) { 
        if ( outputname %in% colnames(SRes) ) {
            col1 <- dt2v( SRes,  outputname )
            ## here we need to acces the indices used in validation
            if ( class(inds)=="list" ) { ## each model has different training indices, access them with name
                valOUT <- Fout[ -inds[which( colnames(Fout)==outputname)], ]
            } else { ## the usual case, training indices are shared across models
                valOUT <- Fout[ -inds, ]
            }
            col2 <- dt2v( valOUT,outputname )
            ## if T then write out column for each model enum and possibly histogram data
            if (write_errors) { 
                dif <- col2 - col1
                Residuals <<- c( Residuals, list(dif) ) ## list encapsulates it into nested level
            }

            RMSE_value <- RMSE(col1, col2)
            RSS_value  <- RSS(col1, col2)
            SAD_value  <- SAD(col1, col2)
            MAD_value  <- MAE(col2, col1)
            MASE_value <- MASE(col2, col1)
            ## AME_value  <- AME(col2, col1)
            
            Surrogate_performance_table[(`output`==outputname &
                                         `method`==current_method &
                                         `preprocessing`==current_preproc),
                                        `RMSE`:= RMSE_value ]
            Surrogate_performance_table[(`output`==outputname &
                                         `method`==current_method &
                                         `preprocessing`==
                                         current_preproc),
                                        `RSS`:= RSS_value ]
            Surrogate_performance_table[(`output`==outputname & `method`==current_method & `preprocessing`==current_preproc),
                                        `SAD`:= SAD_value ]
            Surrogate_performance_table[(`output`==outputname &
                                         `method`==current_method &
                                         `preprocessing`==current_preproc), 
                                        `MAD`:= MAD_value ]
            Surrogate_performance_table[(`output`==outputname & `method`==current_method & `preprocessing`==current_preproc), 
                                        `MASE`:= MASE_value ]
            ## Surrogate_performance_table[(`output`==outputname & `method`==current_method & `preprocessing`==current_preproc), 
            ##                             `AME`:= AME_value ]
            ## Surrogate_performance_table[('output'==outputname & 'method'==current_method & 'preprocessing'==current_preproc), 
            ##     `RMA`:= RMA_value ]
            
            cat("\n",outputname, current_method, current_preproc, "MASE", MASE_value,"MAD",MAD_value)
            ## TODO: Improve feature selection and printing directly from models, this is a start ...
            if ("type" %in% names( model_list[[outputname]] ) ) { ## this is a Dice model
                ## this approach works only on mars, as polymars is different
                if (class(model_list[[outputname]][["model"]])=="mars") {
                    cat(" FS:",mda_mars_importance(model_list[[outputname]])) 
                }
            }
            ## if ("type" %in% names( model_list[[outputname]] ) ) { cat(" FS:",mda_mars_importance(model_list[[outputname]])) }
            ## class(model_list[[1]])
        }
        else if (write_errors) { 
            Residuals <<- c( Residuals, list( rep(NA, nrow(SRes) ) )) ## list encapsulates it into nested level
        }
    })
}

##' @export
WriteModel <- function(write_filepath_prefix,training_samples,seed) {
    ##  ensemble_name <<- paste0( write_filepath_prefix,
    ##                            paste0( c(unique(dt2v(perftop,"method")),unique(dt2v(perftop,"preprocessing")),
    ##                                      training_samples),
    ##                                    collapse="_"),collapse="")
    
    ## OLD saving format: SelectedSurrogate to RDS serialization
    ## saveRDS( SelectedSurrogate, file=paste0(ensemble_name,".RData" ))
  
    ## NEW SAVING FORMAT: FULL IMAGE. Update the sources on import to
    ## overwrite old code in function, otherwise full-backwards
    ## compatability
    
    ## TODO: Some large objects would need to removed from this - like
    ## IO data structures etc otherwise there should be no harm in
    ## saving a full image for reference and execution

    dumpfilename <- paste0(ensemble_name,".RData" )
    cat("\nDumping SelectedSurrogate environment to", dumpfilename,"\n")
    fwrite ( perftop,file= paste0(ensemble_name,"_each_best.csv" ))
    fwrite ( Sperf, file = paste0(ensemble_name,"_all.csv" ))
    ## drop all the redundant large datastructures before we save the image
    fwrite ( Fin,file= paste0(ensemble_name,"_Fin.csv" ))
    fwrite ( Fout, file=paste0(ensemble_name,"_Fout.csv" ))
    ## mrm(c("Fin","Fout","Oin","Oout","Residuals","combined_data","out","all_res","best_res","SRes"))
    save.image( file=dumpfilename )
}

##' @export
WriteModelResiduals <- function(write_filepath_prefix, training_samples, seed) {
  if (!exists(ensemble_name)) {
      ensemble_name <<- paste0( write_filepath_prefix,
                               paste0(format(Sys.time(), "%Y%m%d"), "_"),
                               paste0(c(unique(dt2v(perftop,"method")), unique(dt2v(perftop,"preprocessing")),
                                        training_samples),
                                      collapse="_"),collapse="")
  }
  # for col length safety, extract first the vectors from Residuals list, then data.table
  best_res <<- as.data.table(Residuals[ perftop[,"model_id"] ])
  
  # write names that mean something, so that you know what errors you are looking at 
  colnames(best_res) <- paste ( dt2v( Sperf, "output" ),
                               dt2v( Sperf, "method" ),
                               dt2v( Sperf, "preprocessing" ),
                               sep="_" )[ perftop[,"model_id"] ] # use only the names needed
  ## write all for testing and manual overview
  #fwrite( all_res, "debug/all_residuals.csv" )
  ## write out only for those that have non-zero columns
  fwrite( SelectActiveColumns(best_res), file = paste0(ensemble_name,"_residuals_best.csv" ))
  ## residuals are errors on validation data set - created as inverse from these indexes
  fwrite( as.data.table(inds), file = paste0(ensemble_name,"_train_inds.csv" ) )
  ## err index export
  err_inds = as.data.table( seq(nrow(Fout))[-inds] ) 
  colnames( err_inds ) <- "err_inds"
  fwrite( err_inds, file = paste0(ensemble_name,"_err_inds.csv" ) )
}

## once we have screened all the desired possibilities, we can try to
## select the best combination of surrogate models and their
## pre-processing methods using some kind of global criteria
##' @export
ModelSelector <- function( crit="MASE" ) {
    mrm(c("perftop","Auto_Surrogate_Configuration","avoid_problems_for_sqldf"))
    Sperf[,`RSSQ` := sqrt(RSS)] ## root sum of squares
    Sperf[,`model_id` := seq(nrow(Sperf))]
    ## data table does not have stable inbuilt ids... create a summary
    ## function that includes z-scaled sum of MASE and speed
    Sperf[,`MASE_speed_score` := rowSums(scale(subset(Sperf,select=c(`MASE`,`speed`)),scale=TRUE,center=FALSE))]
    ## so that we can do rbind later with the same cols
    no_aggregation_cols <- ncol(Sperf)
    ## make sure to minimize criteria only for entries that are not
    ## masked with -1
    avoid_problems_for_sqldf <- data.frame( Sperf ) ## causes locked warning that Sperf exists...
    query <- paste0("/* take the best model according to criteria, take fastested in case of overlap */
                  select *,min(",crit,") as min_",crit,
                  " from (select * from avoid_problems_for_sqldf where speed>-1 order by speed asc) as inner group by output ")
                  #order by model_id")
    perftop <<- sqldf(query)
    cat("print query:\n")
    print(query)
    cat("print perftop:\n")
    print(perftop)
 
    ## calculate a multi-criteria z-scaled score for all the models to
    ## faciliate their selection
    speedtop <<- sqldf('select * from Sperf group by output having min(MASE_speed_score)')
    failtop  <<- sqldf('select output,method,preprocessing,method_realname from Sperf where method_realname like "failed%"')
    if ( nrow(failtop)>10 ) {
        cat("\n\nNOTE: Many models failed in fitting!\n\n")
    }
    else if (nrow(failtop)>0) {
        cat("\n\nSOME MODELS FAILED AT FITTING:\n")
        print(failtop)
    } else {
        cat("\nNo models encountered fitting errors!\n")
    }
    ## allows to compare the speed-weigthed model selection to simply
    ## most accurate model selection
    both <- rbind(speedtop[1:no_aggregation_cols],perftop[1:no_aggregation_cols])
    if ("method_realname" %in% colnames(both)) {
        both <- sqldf( "select *,method_realname|| ' ' ||preprocessing as approach from both" )
    } else {
        both <- sqldf( "select *,method|| ' ' ||preprocessing as approach from both" )
    }

    cat("\nMost accurate surrogates by",crit,"for each output:\n")
    print( perftop )
}


## since the number of preprocessor-predictor combinations can range
## into thousands we do not store all the already fitted models, so
## once some desired combination for each output is obtained from
## screening, we rather refit the models additionally, this is a good
## place to enable autotuning for caret models, as those being refit
## are likely already the best in some way, therefore it makes sense
## to apply incremental optimization of tuning parameters only to them
##' @export
Refit <- function( config = perftop, autotune = FALSE, create=TRUE, fitpara = FALSE) {
    ## if (nrow(Sperf)>1) {

    ## iterate on the original order of perftop to get model type,
    ## preprocessing id etc.
    ## assign row.names to config, so that it can be subset correctly when fitting
    row.names(config) <- as.character(config[,"output"])
    cat("\nRefitting the supplied model selection:\n")
    print(config[colnames(Fout),c("method","preprocessing","method_realname")] )  
    ## if the global performance table only has the same number of
    ## tows than model selection table (perftop), then it means there
    ## is only one method with only one
    ## if ( length(unique(as.character(config[,"preprocessing"]))) > 1 ) {
    if (! length(model_list)==nrow(Sperf) ) {## if there was more than just one model type or run, then refit
        mrm( c("Tin","Tout","model_list") )
        FitSurrogates( fitting_config = config,
                      tune = autotune, para=fitpara )
    } else {
        cat("Skipping refit as there was just one model for each output in screening...\n")
    }
    model_sequence <- as.character(config[names(model_list),"method_realname"])
    check_sequence <- GetModelNames()
    preprocessings <- as.character(config[names(model_list),"preprocessing"])

    if ( identical(model_sequence, check_sequence) ) { 
        cat("Model type checks OK!\n")
        if (create) {
            cat("Creating callable function SelectedSurrogate(input)...")
      
            ## Now we create a function that takes a row of parameters
            ## and outputs a row of outputs and save it into the
            ## .GlobalEnv, preprocessing_methods can be specified as a
            ## single method for all outputs or as a vector of
            ## different method for each output the same way the
            ## surrogate methods can also be specified, so every
            ## output the surrogate model needs to predict can use a
            ## different statistical model with different
            ## preprocessing Currently this function is created
            ## manually for this particular example
            SelectedSurrogate <<- function(inp) { Surrogate(inp, preprocessings) }
            cat("Done!\n")
        } else { 
            cat("\nERROR: Model refitting failed!\n\nNO SelectedSurrogate created!\n\n") 
        }
    }
}

##' @export
Surrogate <- function(in_matrix,
                      preprocessing_sequence = rep("none",ncol(Fout)), ## the names of preprocessors for each model in model_list
                      output_model_list = model_list, ## this contains a predictor for each output column
                      immobile_cols = c(),
                      in_rounding_spec = Inf,  ## force rounding of predictions according to named list for each output or the same for all
                      out_rounding_spec = Inf, ## force rounding of predictions according to named list for each output or the same for all
                      ## do_not_clip_negative_to_zero = c("e","pe"), ## concentrations cannot be negative, but some environmental vars can
                      exclude_output_columns = c() ## if specified intigers, then disables prediction for these columns    
) {
    ## predictors take matrices, so make sure it is one
    if (!is.matrix(in_matrix)) { 
        in_matrix <- as.matrix(in_matrix) 
    }

    ## do not pass feature columns into prediction model, that it
    ## hasn't seen before, exclude them if there are any

    ## TODO: fix for self-exclusion!

    if (!in_rounding_spec==Inf) {
        in_matrix <- smartround( in_matrix, in_rounding_spec )
    }
  
    if ( length(output_model_list)<1 )
        stop("ERROR: Output-wise surrogate configuration not found!")
  
    predict_output_cols <- seq(length(names(model_list)))
  
    res <- as.matrix( RunModelList(in_matrix, names(model_list),
                                   output_model_list, 
                                   ## custom_surrogate_inds = predict_output_cols,
                                   inline_preprocessors = preprocessing_sequence) )
  
    ## reapply column names only in the sequence of the outputs actually predicted
    ## colnames(res) <- colnames( Fout )[ predict_output_cols ]
  
    ## we check if any of the immobile columns are actually in the
    ## output match their indices and then reapply to the resuling
    ## matrix
    immobile_inds <- match( immobile_cols, colnames(res) )
    if (length(immobile_inds)>0) {
        print( match( immobile_cols, colnames(Fout)) )
        attr( res, "immobile" ) <- immobile_inds
    }
    ## we save these to enable debugging for simulaton testing
    last_sur_input  <<- in_matrix
    last_sur_output <<- smartround( res, out_rounding_spec )
  
    return( last_sur_output )
}

### histogram density
##' @export
hb <- function() {
    start_up()
    Reshaping(res_full)
    require(histogram)
    ## show me which indexes for this column are larger than 2% of change  
    out_normalized <<- predict.preProcess( preProcess(Oout,"range"), Oout)
    ndl <<- round(Deltas(out_normalized),2)
    ## store sparse indexes for each output in a list
    ## which(abs(round(Deltas(predict.preProcess(predict.preProcess(Oout,"range"),Oout)),2))[[col]]>0.0 )
    sidx <<- llply(colnames(Oout),.fun=function(col) { which( abs(round(Deltas(out_normalized),2))[[col]]>0.02 ) } )
    ## all the sparse output indexes
    allidx <<- unique(unlist(sidx,recursive=T)) ## these were sparse in at least one of outputs
    sparse_for_all <<- sliding_join(sidx)
    ## length(which(abs(round(Deltas(predict.preProcess(prpc[["range"]],Oout)),2))[[8]]==0.0))/nrow(Oout)
}

##' @export
sliding_join <- function(listlike) {
    ## stopifnot(length(listlike)>1,"Sliding join requires that the source iterable has at least 2 elements!")
    ## join the first two
    joinbuff <- intersect(listlike[[1]],listlike[[2]])
    cat(i,length(joinbuff),'\n')
    for (i in seq(3,length(listlike))) {
        joinbuff <- intersect(listlike[[i]],joinbuff)
        cat(i,length(joinbuff),'\n')
    }
    return(joinbuff)
}

## extract table from Marco's Rphree datastructures
##' @export
Reshaping <- function(data) { 
    if (missing(data))
        stop("Reshaping called with no argument!")
### prepare IO data from T and C matrices
### extract original data from RT simulations
    cat('Extracting original data...')
    Oin  <<- TByElem( list_of_elems=data, nested_index=1 ) 
    Oout <<- TByElem( list_of_elems=data, nested_index=2 )
}

### load externally dumped data with autodetection
##' @export
Loadata <- function( IN="dumps/dumpIN.txt", OUT="dumps/dumpOUT.txt" ) { 
    cat('Loading INPUT-OUTPUT tables...')
    Oin <<-  fread(IN)
    Oout <<- fread(OUT)
    
    if ( identical(Oin,Oout) ) 
        stop("FAIL! INPUT-OUTPUT DATA TABLES IDENTICAL!")
}

## This is an idea worth testing - we care about most of these values
## until a certain level of digits. However, model drift can cause
## some of the values to be out of input ranges that the prediction
## model is able to deal with reasonably well. The default option
## would be to resampl the simulator in those regions. A quicker
## heuristic also worth trying is to round the tables to coarser
## "resolution" of numbers that are meaningful for the scenario and
## fit those as alternative "extrapolating" options.
##' @export
MultiRound <- function( t, rounding_digits=c(8,7,6) ) {
    out_table <- t
    for (digits in rounding_digits) { out_table <- rbind( out_table, smartround(t, digits) ) }
    return ( out_table )
}

## additional cleaning operations
##' @export
Filtering <- function(
                      roundto=Inf, ## setting Inf switches rounding off
                      dedup=TRUE,
                      exclude=numeric(),
                      exclude_output=c(),
                      try_multi_round = c(),
                      to_factor = FALSE, ## convert columns with low number of unique valus to factors?
                      ## F - all neg forced to zero
                      ## T - all allowed neg
                      ## c("field","names") - allow only these cols to be negative
                      allow_negative = T#c("Pressure","Charge")
                                        #exclude=6
                      )
{
    ## if negatives are forced to zero then we start with that
    if ( class(allow_negative)=="logical" ) {
        if (!allow_negative) { ## of F then all are forced to 0
            cat('NEGATIVES: Clipping all negative values in all data to zero...be sure this is really what you want!!!')
            Oin[Oin<0]=0
            Oout[Oout<0]=0 
        }
    } else { ## if field names are specified then the rest are also clipped to zero
        Oin <<- AllowSomeNegativeColumns( Oin, allow_negative)
        Oout<<- AllowSomeNegativeColumns( Oout,allow_negative)
    }
  
    ## make sure there is actually data, if none of the filtering options are actually run
    Fin <<- Oin
    Fout<<- Oout
  
    ## deduplication works by syncing indexes in the two tables
    cat('excluding...')
    if ( length(exclude) > 0 ) {
                                        #Oin  <<- subset(Oin, select = setdiff(colnames(Oin), exclude) )
                                        #Oout <<- subset(Oout,select = setdiff(colnames(Oout),exclude) )
        Fin  <- exc(Oin, exclude)
        Fout <- exc(Oout,exclude)
    }
    
    cat('also outputs',exclude_output,"...")
    if ( length(exclude) > 0 ) {
        Fout <- exc(Fout,exclude_output)
    }
    
    if (!roundto==Inf) {
        cat("\nrounding...")
        Fin  <- smartround(  Fin,roundto )
        Fout <- smartround( Fout,roundto )
    }
    
    if ( length(try_multi_round) > 0 ) {
        cat("multirounding...")
        Fin  <- MultiRound(  Oin,try_multi_round )
        Fout <- MultiRound( Oout,try_multi_round )
    }
    
    cat("checking for static input columns...")
    Fin  <- SelectActiveColumns(Fin)
    Fout <- SelectActiveColumns(Fout)
    
    ## remove non-unique parameter combinations after filtering and
    ## rounding
    if (dedup) {
        cat("deduplicating...")
        dups <<- duplicated(Fin)
        Fin  <<- Fin[ which(dups==F), ] 
        Fout <<- Fout[ which(dups==F), ]
        redundancy=(sum(dups==T)/nrow(Oin))*100
        cat( redundancy,"%...")
    }
    
  
    if (to_factor) {
        cat("looking for OUTPUT columns that could be converted to factors for using classification instead of regression...\n")
        Fout <<- Cols2Fact(Fout)
    }
  
    cat("IN-OUT different?...")
    n_the_same <- NumericallyCompareTables(Oin,Oout)
    n_common_colnames <- length( intersect(colnames(Oin),colnames(Oout)) )
    if ( n_the_same/n_common_colnames > 0.8 & n_common_colnames>1 )
        stop (" ERROR: TOO MANY IDENTICAL IN-OUT COLUMNS, CHECK YOUR DATA!\n")
    cat('done!\n')
}

##' @export
AllowSomeNegativeColumns <- function( dfrmlike, allow_negative_cols ) {
  if ( length(allow_negative_cols)>0) {
      neg_cols_actually_exist = allow_negative_cols[ allow_negative_cols %in% colnames(dfrmlike) ]
  }

  if ( length(neg_cols_actually_exist) > 0 ) {
      temp1 <- subset(dfrmlike, select = neg_cols_actually_exist) ## prevent <0   
  }

  temp2 <- subset(dfrmlike, select = setdiff(colnames(dfrmlike),allow_negative_cols) )
  temp2[temp2<0]<-0
  if ( length(neg_cols_actually_exist) > 0 ) { allcols <- cbind(temp2,temp1) }
  else allcols <- temp2 
  return(allcols)
} 

##' @export
Relcal <- function() {
    ## only do this on the performance table columns that are numerics and therefore suitable
    relacols <<- colnames(Sperf)[laply( Sperf,.fun=class )=="numeric"]
    ## todo: finish (round(colwise_rescale(subset(Sperf,select=relacols),minval=0,maxval=1),5)-1)*-1

    llply( relacols, .fun=function(colname) { 
        colvals <<- dt2v(Sperf,colname)
        ## cat(colname,colvals/sort(colvals)[1],"\n")
        round(0.0e-32 - ( colvals / max(colvals))*-1,5)
        ## Sperf[ paste0("relat_",colname):=colvals/min(colvals) ]
    })
}



## caret has own pre-processing infrastructure but not everything does
## also, we may wish to experiment a little with experiment-specifc
## preprocessing
##' @export
SelectPreProc <- function(method=8, IOranges=originalIOranges, v=TRUE)
    ## if this is passed in, that means we need
    ## to preprocess it using previously stored
    ## ranges given_input=NULL
{
    mrm( c("PPC","PPCout"), verbose=FALSE )
    if (v)
        cat('\nApplying preprocessing with',specific_preprocessing_methods[method],'...')
    
    ## Here we create the preprocessing method as a function that can
    ## be uniformly called globally. Most often only input may need to
    ## be normalized, but sometimes some of the approaches may require
    ## the optional original range parameters (original min max in
    ## case of stretch and center and scale in case of z-scaling) this
    ## is very imporant because when the surrogate model is used
    ## instead of simulation model and gets a table of inputs for
    ## which it needs to predict the simulator outputs, its
    ## normalization has to use the same parameters that where used
    ## for the original training data, as otherwise the predictor
    ## actually gets input that is different than what we may think it
    ## gets, causing wrong predictions for purely technical reasons.
  
    ## Overall there are some general ways to use preprocessing
    ## 1. preprocess only inputs, usually scale something using given input and output ranges
    ##    ==> defines PPC function for use on input tables
    ## 2. preprocess also output table in some way unrelated to the input table 
    ##    ==> defines PPCout function and PPCback backtransform functions 
    ##    ==> defines PPCout function for use on input tables
    ## 3. preprocess both tables using some formula between themselves (like substract input from output) 
    ##    this is complicated, costly and possibly wrong, TODO: need to discuss
    ## 4. do nothing :) - sometimes best as some methods actually like this:)   

    if (method==1) {
        PPC <<- function(inp){ rescaling( Deltas(inp), IOranges,  0,1)}
        PPCout <<- function(inp,outp) {rescaling( Deltas(outp),IOranges,  0,1)}
    }
    
    if (method==2) {
        PPC <<- function(inp)      { rescaling( Deltas(inp), IOranges, -1,1) }
        PPCout <<- function(inp,outp) { rescaling( Deltas(outp),IOranges, -1,1) }
    }
    if (method==3) {
        PPC <<- function(inp) { scale( Deltas(inp) ) }
    }
  ##                PPCout <<- function(inp,outp) { rescaling( Deltas(outp),IOranges, -1,1) }}
    if (method==4) {
        PPC <<- function(inp) { scale( inp ) }
    }
    if (method==5) {
        PPC <<- function(inp) { Deltas( scale(inp) ) }
    }
    if (method==6) {
        PPC <<- function(inp) { Deltas( rescaling(inp, 0,1) ) }
        PPCout <<- function(inp,outp) { Deltas( rescaling(outp, IOranges,0,1)) }
    }
    if (method==7) {
        PPC <<- function(inp) { Deltas( rescaling(inp, IOranges,-1,1) ) }
        PPCout <<- function(inp,outp) { Deltas( rescaling(outp,-1,1) ) }
    }
    if (method==8) {
        PPC <<- function(inp) { rescaling( inp, IOranges,   0,1)   }
    }
    if (method==9) {
        PPC <<- function(inp) { rescaling( inp, IOranges,   0,10)  }
    }
    if (method==10){
        PPC <<- function(inp) { rescaling( inp, IOranges, -10,10)  }
    }
    if (method==11){
        PPC <<- function(inp) { colwise_rescale( Deltas( inp ),  0,10) }
    }
    if (method==12){
        PPC <<- function(inp) { colwise_rescale( Deltas( inp ),-10,10) }
    }
    if (method==13){
        PPC <<- function(inp){ rescaling( inp, IOranges, 0,1) }
        PPCout <<- function(inp,outp) { rescaling( outminus(inp,outp), IOranges, 0,1) }
    }
    if (method==14){
        PPC <<- function(inp)      { rescaling( inp, IOranges, 0,10) }
        PPCout <<- function(inp,outp) { rescaling( outminus(inp,outp), IOranges, 0,10) }
    }
    if (method==15){
        PPC <<- function(inp)      { rescaling( inp, -10,10)}
        PPCout <<- function(inp,outp) { rescaling( outminus(inp,outp), IOranges, -10,10) }
    }
    if (method==16){
        PPC <<- function(inp)      { scale( inp ) }
        PPCout <<- function(inp,outp) { scale( outminus(inp,outp) ) }
    }
    if (method==17){ ## doing nothing may also be worth chekcing out :)
        PPC <<- function(inp) { inp }
    }
    
    ## somewhat confusingly the predict call takes previously stored
    ## pre-processing transform to apply to supplied data also, here
    ## we create a function PPC that will actually do the transform
    ## when needed!
    if (method==18){
        PPC <<- function(inp) { predict(original_preproc[["pca"]],inp) }
    }
    if (method==19){
        PPC <<- function(inp) { predict(original_preproc[["ica"]],inp) }
    }
    if (method==20){
        PPC <<- function(inp) { predict(original_preproc[["BoxCox"]],inp) }
    }
    if (method==21){
        PPC <<- function(inp) { predict(original_preproc[["YeoJohnson"]],inp) }
    }
    if (method==22){
        PPC <<- function(inp) { predict(original_preproc[["expoTrans"]],inp) }
    }
    if (method==23){
        PPC <<- function(inp) { predict(original_preproc[["range"]],inp) }
    }
    if (method==24){
        PPC <<- function(inp) { predict(original_preproc[["scale"]],inp) }
    }
    if (method==25){
        PPC <<- function(inp) { predict(original_preproc[["spatialSign"]],inp) }
    }
    if (method==26){
        PPC <<- function(inp) { predict(original_preproc[["c(\"center\", \"scale\")"]],inp) }
    }
}

##' @export
NumericallyCompareTables <- function( in_table, out_table ) {
    common_cols <- intersect(colnames(in_table),colnames(out_table) )
    table_diff  <- sapply(common_cols,
                          FUN=function(col) { 
                              coldiff <- dt2v(in_table,col)-dt2v(out_table,col)
                              any(coldiff > 0)})
    cat("Completely identical IN and OUT columns =",sum(table_diff==FALSE),":",
        names(table_diff)[table_diff==FALSE],"\n")
    return(sum(table_diff==FALSE))
}

##' @export
SubstractCommonTableCols <- function(tab1,tab2) {
    common_cols <- intersect(colnames(tab1),colnames(tab2) )
    sapply(common_cols, FUN=function(col) { dt2v(tab1,col)-dt2v(tab2,col) } )
}

## Creates Tin and Tout functions for 
##' @export
BatchPreProc_ <- function() {
    ## if special output preprocessor function then uses dependencies
    ## beteen IO is defined, then use it
    if ( exists("PPCout") ) { 
        cat('\nSpecial output preprocessing used!\n')
        Tout <<- data.matrix( PPCout(Fin, Fout ) ) 
    } else { ## otherwise just use the filtered output
        Tout <<- data.matrix( Fout )
    }  ## if not than just make sure it is a matrix
    Tin <<- PPC(data.matrix( Fin ) )
}

## Creates Tin and Tout functions for 
##' @export
BatchPreProc <- function(selcol=c()) {
    ## if special output preprocessor function then uses dependencies
    ## beteen IO is defined, then use it
    if ( length(selcol)>0 ) { 
        cat("Selecting",selcol,"from output\n")
        selout <- dt2v( Fout, selcol ) 
    }
    else {
        selout <- Fout
    }
    
    if ( exists("PPCout") ) {
        cat('\nSpecial output preprocessing used!\n')
        Tout <<- data.matrix( PPCout( Fin, selout ) )
    }
    else { ## otherwise just use the filtered output
        Tout <<- data.matrix( selout )
    }  ## if not than just make sure it is a matrix
    Tin <<- PPC( data.matrix(Fin) )
}

## This is something critically important: when we get the overall
## original dataset in the beginning, we have to store initial ranges
## and parameters for all preprocessing options that could be used.
## This is because we will want to re-use outputs again as inputs and
## when they are passed again back to the predictor. Also, if values
## outside the original ranges get passed to the predictor, the
## normalization basis should not be confused by this. This will not
## adversly affect range, scale and the usual preprocessors, but could
## adversly affect distribution based methods!
##' @export
InitPreProc <- function(dfrm,
                        pre_proc_methods_inds=c(17),
                        ## caret_methods=list("scale")#,"range","BoxCox",#"expoTrans",#"YeoJohnson","pca","spatialSign",
                        ## c('center','scale')),
                        para=TRUE)
{
    cat("\nInitilizing preprocessors...")
    preproc_names <- specific_preprocessing_methods[pre_proc_methods_inds]
    caret_methods <<- sapply(preproc_names,FUN=function(ppn) {
        out <- ppn
    
        if (ppn=="center and scale")
            out <- c('center','scale')  
        if (ppn=="smart")
            out <- c('zv','nzv','conditionalX') 
    
        return( out ) }, simplify = TRUE, USE.NAMES=TRUE)
    
    if ( is.na(caret_methods[1]) )
        caret_methods <<- caret_methods[-1]
    
    if (("no" %in% names(caret_methods)) & length(caret_methods)>1 ) { 
        caret_methods <<- caret_methods[2:length(caret_methods)] 
        caret_methods <<- as.character(caret_methods)
        print(caret_methods)
        preproc <- llply(caret_methods,.fun=function(m) { preProcess(dfrm,m) },.parallel=para)
        names(preproc) <- caret_methods
    
    } else { preproc = list() }
  
    return(preproc)
}

## this does generally wrap the caret downsampling capabilities, but
## tries to do the same for regression by casting the outputs as char
## classes
##' @export
DownSampler <- function( ins,outs ) {
    ur <- coluniratio(outs)
    cat("\nOutput column has",ur*100,"% unique values.\nWill downsample the data now...")
    cat("\ncasting to char...")
    clasouts <- as.factor(as.character(as.matrix(outputs)))
    cat("downsampling...")
    dr <- downSample(inputs,clasouts,list=T)
    cat("got",nrow(dr$x),"downsampled samples!")
    return(dr)
}

##' @export
OverrideValueRange <-function( intable, minval=0.23, maxval=0.23, newval=0.0981 ) {
    cat("\nConstraining value range to")
    intable[intable<=minval] <- newval
    intable[intable>=minval] <- newval
    cat("\nWill use",length(after_outliers_idxes),"rows in training")
}



## =========================== TRAIN ====================
## Run the full cycle surrogate creation and test macro

## - it tries out a bunch of predefined statistical models or all the available, if not specified
## - together with specified preprocessing methods (see header for method enum descriptions)
## - and column rounding specification. the rounding spec is important because otherwise You 
##   may end up trying to predict numerical effects in the model output instead of meaningful output

##' @export
Train <- function(surrogate_types = c(),
                  preprocessing_ind = c(17,22,23,24,25,26), ## set this to actual data str to supply loaded tables
                  input_data = "dumps/amd_course_IN.txt",
                  output_data = "dumps/amd_course_OUT.txt", seed = 105, ## select smallest by this criteria
                  selection_criteria = "MASE", ## fraction of rows from training data
                  training_samples=0.25, ## TODO: MULTIPASS ACTIVE
                  ## LEARNING works when training_samples < 0.5,
                  ## better with 0.45 you can set unlimited number of
                  ## passes for this what it does is trains the first
                  ## pass with assigned ratio of training samples,
                  ## runs two-fold validation, then it takes another
                  ## slice the size of training_samples composed of
                  ## indices with largest errors, trains again as many
                  ## passes as you have set until stopping threshold
                  ## reached Idea is very simple: include data in
                  ## training that the model struggles with
                  multipass_training = 0, ## no more than this goes to multipass training data
                  mutlipass_stopping_threshold = 0.8, ## a quick way to disable models that break current exeriment
                  surrogate_blacklist = c(), ## caret preprocessors are mostly better so the old ones are out
                  
                  preprocessing_blacklist = c(3,5,11,12,13,14,15,16,19),
                  exclude_columns = c(),
                  ## this is specific to coupled chemistry surrogates -
                  ## columns here will not be part of ensemble and will
                  ## not predicted in output, but will be given as
                  ## inputs in validation. the reasoning is that these
                  ## parameters come from the couple model (flow?) and
                  ## are thus external to chemistry - are therefore
                  ## forced on the surrogate completely from the
                  ## outside
                  external=c("Time","Pressure"),
                  allow_neg_cols=c("pe"), #,multiround = c(9,6)
                  multiround = c(), #,exclude_output_columns = c("e")
                  train_para=TRUE, run_para=FALSE, preproc_para=TRUE,
                  use_cores = 0 ,tuner = FALSE, ## you can apply a rounding specification, but it usually does more harm than good
                  r_digits = Inf, write_model_data=TRUE,
                  session_path = FALSE,
                  write_full_residuals=FALSE,
                  write_filepath_prefix="models/"
                  )
{  	  
    ## run only the preprocessing methods that have not been blacklisted
    preprocessing_ind <- setdiff( preprocessing_ind, preprocessing_blacklist )
    
    start_up( session_path = session_path, use_cores=use_cores) ## load libraries etc
    
    ## if ( use_cores>0 ) { 
    ##     if ( getDoParRegistered() ) {
    ##         cat("\nManually forcing",use_cores,"cores for parallelization!\n")  
    ##         registerDoMC( use_cores ) 
    ##     }
    ## } else { cat("\nTried to set number of processing cores manually, but doParallel does not appear to be loaded, ignoring!\n\n") } 
    
    ## cleanup, so no old data strs can affect new batch by mistake
    mrm( c("Tin","Oin","Oout","Cout","Fin","Fout","Tout","SRes","Sperf","valOUT","perftop","Residuals") )
    histograms <<- list()
    Residuals  <<- list()
    ## assign globally if actual data structure passed
    if ("data.frame" %in% class(input_data) & "data.frame" %in% class(output_data) ) {
        cat("Setting input and output data from existing data structures...\n")
        Oin <<- input_data
        Oout<<-output_data
    } else if ( length(input_data)>0 & length(output_data)>0 & class(input_data)=="character" & class(output_data)=="character") { 
        cat("\nLoading input output data from", input_data, output_data, "\n")
        Loadata( IN=input_data, OUT=output_data)
    }
    ## we basically do this here because we need the names for output columns to create
    else { 
        cat("\n Attempting to read data from Rphree simultation structures...")
        Reshaping( res_sim ) 
    }
    
    cat("\nCorrelations in original data:\n")
    Fcor <- cor(Oin,Oout)
    print( Fcor )
    
    Filtering( roundto = r_digits, exclude=exclude_columns, 
              exclude_output=external, try_multi_round = multiround,
              allow_negative=allow_neg_cols) ## apply rounding and deduplication for training
    
    ## here access columns in input data, so that evaluation table and
    ## tunegrids could be created tuning parmaters for some methods
    ## are sometimes dependent on the number of source outputs and
    ## considerations that are easier to set when the data has already
    ## been loaded, therefore we call it only here, after Reshaping
    ## and Filtering
    
    ## stop("Breakpoint reached")
    createTuneGrids()
    
    ## if no surrogate method list was given, throw everything that we've got
    if ( length(surrogate_types)==0 ) { 
        surrogate_types = c("mda_p01","polymars_p01","rpart_anova_direct",
                            ## setting up basic tuning parameters is equivalent 
                            ## to enabling the caret method for use globally
                            names(multiTuneGrids),
                            DiceMethods)
    }
    ## remove models that have been blaclisted and make sure no
    ## duplicate models are ran, if passed by accident, through a set
    ## op. Also, we run all of the models if no specific method list
    ## is given
    surrogate_types <- setdiff( surrogate_types, surrogate_blacklist )
    
    ## Create perfomance table with rows for each of outputs 
    Sperf <<- data.table( expand.grid(output=colnames(Fout), method=surrogate_types, 
                                      preprocessing=specific_preprocessing_methods[preprocessing_ind]) )
    ## create columns for evaluation
    Sperf[,`speed`:=rep(-1.0,nrow(Sperf))] ## negative sets correct col datatype, but meaning is like NA
    
    ## WARNING: do not access combined_data and try to use it. It is
    ## not the full IO data You should use.
    
    ## NOTE: regarding IO ranges, if the same variables are present
    ## both in the input and output then scaling is performed on their
    ## combined range. This is useful as the values ranges present in
    ## input and output can vary markedly. Thus preserving their
    ## combined initial ranges allows to avoid having to deal with
    ## input and output scaling separately as the same original range
    ## can be used for both. This is somewhat specific to recursive
    ## use of geochemistry surrogates.
    if ( length( intersect(colnames(Fin),colnames(Fout)) ) > 0 ) {
        combined_data <<- rbind( Fin, Fout, fill=T )
        comb_len <- nrow(combined_data)
        combi_dups <- duplicated(combined_data)
        combined_data <<- combined_data[ which(combi_dups==F), ]
        combi_red <- (sum(combi_dups==T)/nrow(combined_data))*100
        cat("\nCombined IN-OUT redundancy:",combi_red,"\n")
        originalIOranges <<- GetRanges( combined_data )
        ## if no preprocessin is used and it is the only thing then do not store anything
    
        if ( !(length(preprocessing_ind)==1 & preprocessing_ind[1]==17)  ) {
            original_preproc <<- InitPreProc( combined_data, pre_proc_methods_inds = preprocessing_ind, para = preproc_para)
        }
    } else {
        ## if there are no common columns in Fin and Fout, then we just normalize Fin
        originalIOranges <<- GetRanges( Fin )
        original_preproc <<- InitPreProc( Fin, pre_proc_methods_inds = preprocessing_ind,  para = preproc_para)
    }
  
    ## originalIOranges <<- GetRanges( Fin )
    cat("Percentages of unique values in columns of input and output data:\n")
    print( uvt(Fin,Fout) )
    cat("\nNormalization ranges:\n")
    print( originalIOranges )
    
    ## Cast to matrix only after hi-speed init operations are done!
    ## Fin <<- data.matrix(Fin)
    ## Fout <<- data.matrix(Fout)
    
    ## take a random sample amounting to given % of available traning data
    ## inds <<- sample( nrow(Tin), nrow(Tin)*training_samples )
    
    ## THIS HAS TO BE HERE - it makes sure that the same samples are
    ## used for all model training unless multipass is specified. In
    ## case of multipass it allows to preserve the training sequence
    ## from the same initial state for all. In case of randomly
    ## sampled two-fold cross validation this is it - Run and Fit take
    ## these from .GlobalEnv and resample the data
    set.seed( seed )
    cat("\nRandom sampling",round(training_samples*100,2),"% of data, seed",seed,"\n")
    inds <<- rsi( nrow(Fin), training_samples )
    gc() ## force garbage collector on everything that came before
  
    ## START WORKFLOW HERE
    for (i in preprocessing_ind) {
        current_preproc <<- specific_preprocessing_methods[i]
        SelectPreProc( originalIOranges, method=i )
        ## BatchPreProc() ## create Tin and Tout as training in-out data tables
        for (surrogate_type in surrogate_types) {
            current_method <<- surrogate_type
            
            ## every surrogate model with every preprocessing step
            ## will have different training samples for input and
            ## response tables and the old result table is always
            ## removed to make it very unlikely that old R data
            ## structures are preserved and lead to misleading results
            mrm( c("valOUT","valIN","SRes","model_list") ) ## remove data structures just to be sure
            if (identical(Fin,Fout))
                stop("ERROR: Input and output traning data is the same!")
            
            ## Fit the randomly sampled indices
            cat("\nFitting ",surrogate_type,"\n")
            FitSurrogates( surrogate_type, para=train_para, tune = tuner )
            cat("FitSurrogates ended\n")
            ##            gc()

            ## try using them for prediction
            RunSurrogates( method = surrogate_type )
            cat("RunSurrogates ended\n")
            ## it is important to round the results to the number of decimal 
            ## digits that make them physically meaningful, for this we re-used the 
            ## originally supplied rounding specification
            
            SRes <<- smartround( SRes, roundto = r_digits )
            ## Evaluate model errors and make for those columns that
            ## are both in the predictor output and the validation
            ## part of the output
            write_evals( ) ## only for those having both original values and resulting value
            ## start multi-pass training here
        }
    }

    ## this call creates a surrogate spec in form of a dataframe where
    ## rows are outputs and cols are the selected properties for the
    ## best approximation of the particular output
    ModelSelector( selection_criteria )
    
    ## This refits the selected surrogate model configuration after
    ## the screening runs and evaluation in the ModelSelector
    Refit( config=perftop, fitpara=train_para )

    ## we share this for file names in saving function ahead
    ensemble_name <<- paste0(write_filepath_prefix,
                             paste0( c(unique(dt2v(perftop,"method")),unique(dt2v(perftop,"preprocessing")),
                                       training_samples),
                                    collapse="_"),collapse="")
  
  
    if (write_full_residuals) 
        WriteModelResiduals(write_filepath_prefix,training_samples,seed) 
    if (write_model_data) 
        WriteModel(write_filepath_prefix,training_samples,seed)  
}
